from typing import List, cast

from fastapi.routing import APIRoute

from pest.logging import log

from ..core.application import PestApplication
from ..core.module import setup_module
from ..core.types.fastapi_params import FastAPIParams
from ..metadata.meta import get_meta_value
from ..middleware.types import MiddlewareDef


def make_app(
    fastapi_params: FastAPIParams,
    root_module: type,
    middleware: MiddlewareDef = [],
    prefix: str = '',
) -> PestApplication:
    """Creates the pest application instance"""
    module_tree = setup_module(root_module)
    routers = module_tree.routers
    app = PestApplication(module=module_tree, middleware=middleware, **fastapi_params)

    for router in routers:
        # log routes while we add em
        log.info(f'Setting up {get_meta_value(router, "name")}')
        for route in cast(List[APIRoute], router.routes):
            for method in route.methods:
                full_route = f'{prefix}{route.path}'

                if full_route == '/' or not full_route.endswith('/'):
                    log.debug(f'{method: <7} {full_route}')

        # add the router
        app.include_router(router, prefix=prefix)

    return app



import logging
from datetime import time, timedelta
from enum import Enum
from multiprocessing.context import BaseContext
from typing import (
    TYPE_CHECKING,
    Callable,
    List,
    TextIO,
    Tuple,
    TypedDict,
    Union,
)

LOGGER_NAME = 'pest'

if TYPE_CHECKING:
    try:
        from loguru import (
            CompressionFunction,
            FilterDict,
            FilterFunction,
            FormatFunction,
            Message,
            PathLikeStr,
            Record,
            RetentionFunction,
            RotationFunction,
            Writable,
        )
    except ImportError:
        pass


class LogLevel(int, Enum):
    """🐀 ⇝ logging levels"""

    FATAL = 50
    ERROR = 40
    WARN = 30
    INFO = 20
    DEBUG = 10
    NONE = 0


log = logging.getLogger(LOGGER_NAME)
log.setLevel(LogLevel.DEBUG)


class SinkOptions(TypedDict, total=False):
    """🐀 ⇝ config options for loguru sinks"""

    sink: Union[
        str, 'PathLikeStr', TextIO, 'Writable', Callable[['Message'], None], logging.Handler
    ]
    level: Union[str, int]
    format: Union[str, 'FormatFunction']
    filter: Union[str, 'FilterFunction', 'FilterDict']
    colorize: bool
    serialize: bool
    backtrace: bool
    diagnose: bool
    enqueue: bool
    context: Union[str, BaseContext]
    catch: bool
    rotation: Union[str, int, time, timedelta, 'RotationFunction']
    retention: Union[str, int, timedelta, 'RetentionFunction']
    compression: Union[str, 'CompressionFunction']
    delay: bool
    watch: bool
    mode: str
    buffering: int
    encoding: str


class LoggingOptions(TypedDict, total=False):
    """🐀 ⇝ config options for loguru"""

    intercept: List[Union[str, Tuple[str, LogLevel]]]
    '''list of built-in loggers to intercept using `loguru`'''
    shush: List[str]
    '''list of built-in loggers to shush using `loguru`'''
    level: LogLevel
    '''default log level for all loggers'''
    format: Union[str, Callable[['Record'], str], None]
    '''override default log format'''
    access_log: bool
    '''whether to enable access logging'''
    verbose: bool
    '''enables all loggers'''
    sinks: List[SinkOptions]



from inspect import getmembers, isfunction
from typing import TYPE_CHECKING, Any, ClassVar, List, Optional

from fastapi.routing import APIRoute

from ..exceptions.base.pest import PestException
from ..metadata.meta import get_meta, get_meta_value, inject_metadata
from ..metadata.types._meta import PestType
from ..metadata.types.controller_meta import ControllerMeta
from ..metadata.types.handler_meta import HandlerMeta
from ..utils.fastapi.router import PestRouter
from ..utils.functions import classproperty
from .common import PestPrimitive
from .handler import HandlerTuple, setup_handler
from .types.status import Status

if TYPE_CHECKING:  # pragma: no cover
    from .module import Module

NOT_CONTROLLER = 'Class {cls} is not a subclass of Controller'
NOT_ROUTER = 'Object is not a subclass of PestRouter'


def setup_controller(cls: type, module: Optional['Module'] = None) -> None:
    """🐀 ⇝ sets up a `controller` class"""
    if not issubclass(cls, Controller):
        raise PestException(NOT_CONTROLLER.format(cls=cls.__name__))
    cls.__setup_controller_class__(module)


def router_of(cls: type) -> PestRouter:
    """🐀 ⇝ obtains a `controller`'s router"""
    if not issubclass(cls, Controller):
        raise PestException(NOT_CONTROLLER.format(cls=cls.__name__))

    if cls.__router__ is None:
        raise PestException(f'Controller {cls.__name__} has not been setup')

    return cls.__router__


def routes_of(cls: type) -> List[APIRoute]:
    """🐀 ⇝ obtains all routes of a `controller`"""
    return router_of(cls).routes


def module_of(cls: type) -> 'Module':
    """🐀 ⇝ obtains the parent module of a `controller`"""
    if not issubclass(cls, Controller):
        raise PestException(NOT_CONTROLLER.format(cls=cls.__name__))

    mod = cls.__parent_module__
    if mod is None:
        raise PestException(f'Parent module of controller {cls.__name__} has not been setup')

    return mod


class Controller(PestPrimitive):
    __router__: ClassVar[PestRouter]
    __parent_module__: ClassVar[Optional[Any]]

    @classproperty
    def __pest_object_type__(cls) -> PestType:
        return PestType.CONTROLLER

    @classmethod
    def __str__(cls) -> str:
        meta = get_meta(cls, ControllerMeta, clean=True)
        return f'Controller {meta.prefix}'

    @classmethod
    def __setup_controller_class__(cls, module: Optional['Module']) -> None:
        """sets up a controller class"""
        meta = get_meta(cls, dict, clean=True)
        cls.__router__ = PestRouter(**meta)
        cls.__parent_module__ = module
        inject_metadata(cls.__router__, name=f'{cls.__name__} {meta.get("prefix", "")}')

        router = cls.__make_router__()
        cls.__router__.include_router(router)
        cls.__class_status__ = Status.READY

    @classmethod
    def __make_router__(cls) -> PestRouter:
        """sets up a controller's router"""
        r = PestRouter()

        for route in cls.__routes__():
            r.routes.append(route)

        return r

    @classmethod
    def __routes__(cls) -> List[APIRoute]:
        """makes routes for the controller"""

        routes = []
        for handler in cls.__handlers__():
            ruote = setup_handler(cls, handler)
            routes.append(ruote)
        return routes

    @classmethod
    def __handlers__(cls) -> List[HandlerTuple]:
        members = getmembers(cls, lambda m: isfunction(m))
        handlers: List[HandlerTuple] = []

        for _, method in members:
            meta_type = get_meta_value(method, key='meta_type', type=PestType, default=None)
            if meta_type == PestType.HANDLER:
                handlers.append((method, get_meta(method, HandlerMeta)))

        return handlers



from abc import ABC, abstractmethod
from inspect import isclass
from typing import Union

from ...utils.functions import classproperty

try:
    from typing import TypeGuard
except ImportError:
    from typing_extensions import TypeGuard

from pest.metadata.types._meta import PestType

from ...exceptions.base.pest import PestException
from ..types.status import Status


class PestPrimitive(ABC):
    __class_status__: Status = Status.NOT_SETUP

    @classproperty
    @abstractmethod
    def __pest_object_type__(cls) -> PestType:
        """🐀 ⇝ returns the type of pest object"""
        ...


def is_primitive(clazz: Union[type, object]) -> TypeGuard[PestPrimitive]:
    """🐀 ⇝ checks if a class is a pest primitive"""
    if not isclass(clazz):
        return isinstance(clazz, PestPrimitive)

    return issubclass(clazz, PestPrimitive)


def status(clazz: Union[type, object]) -> Status:
    """🐀 ⇝ gets a primitive's status"""
    is_class = isclass(clazz)
    name = clazz.__name__ if is_class else clazz.__class__.__name__
    if not is_primitive(clazz):
        raise PestException(f'{name} is not a pest primitive.')

    if is_class:
        return getattr(clazz, '__class_status__', Status.NOT_SETUP)

    return clazz.__class_status__


def primitive_type(clazz: Union[type, object]) -> PestType:
    """🐀 ⇝ gets a primitive's type"""
    is_class = isclass(clazz)
    name = clazz.__name__ if is_class else clazz.__class__.__name__

    if not is_primitive(clazz):
        raise PestException(f'{name} is not a pest primitive.')
    if is_class:
        return getattr(clazz, '__pest_object_type__')

    return clazz.__pest_object_type__



from typing import (
    Any,
    Callable,
    Coroutine,
    Dict,
    List,
    Sequence,
    Type,
    TypedDict,
    Union,
)

from fastapi import FastAPI, Request, Response
from fastapi.params import Depends
from fastapi.routing import APIRouter
from starlette.routing import BaseRoute
from starlette.types import Lifespan


class FastAPIParams(TypedDict, total=False):
    debug: bool
    routes: List[BaseRoute]
    title: str
    summary: str
    description: str
    version: str
    openapi_url: str
    openapi_tags: List[Dict[str, Any]]
    servers: List[Dict[str, Union[str, Any]]]
    dependencies: Sequence[Depends]
    default_response_class: Type[Response]
    redirect_slashes: bool
    docs_url: str
    redoc_url: str
    swagger_ui_oauth2_redirect_url: str
    swagger_ui_init_oauth: Dict[str, Any]
    exception_handlers: Dict[
        Union[int, Type[Exception]],
        Callable[[Request, Any], Coroutine[Any, Any, Response]],
    ]
    on_startup: Sequence[Callable[[], Any]]
    on_shutdown: Sequence[Callable[[], Any]]
    lifespan: Lifespan[FastAPI]
    terms_of_service: str
    contact: Dict[str, Union[str, Any]]
    license_info: Dict[str, Union[str, Any]]
    openapi_prefix: str
    root_path: str
    root_path_in_servers: bool
    responses: Dict[Union[int, str], Dict[str, Any]]
    callbacks: List[BaseRoute]
    webhooks: APIRouter
    deprecated: bool
    include_in_schema: bool
    swagger_ui_parameters: Dict[str, Any]
    separate_input_output_schemas: bool



from typing import List, Sequence, Type, TypedDict, Union

try:
    from typing import TypeAlias
except ImportError:
    from typing_extensions import TypeAlias

from starlette.middleware import Middleware as StarletteMiddleware

from .base import PestMiddlwareCallback

MiddlewareDef: TypeAlias = Sequence[
    Union[StarletteMiddleware, Type[PestMiddlwareCallback], PestMiddlwareCallback]
]


class CorsOptions(TypedDict, total=False):
    allow_origins: List[str]
    '''
    Configures the `Access-Control-Allow-Origin`.

    Expects a list of origins that should be allowed to make cross-origin requests, or `"*"`
    to allow all origins.
    '''
    allow_methods: List[str]
    '''
    Configures the `Access-Control-Allow-Methods` header.

    Expects a list of allowed HTTP methods as strings, e.g. `["GET", "POST", "PUT", "DELETE"]`.
    Defaults to `["GET"]`
    '''
    allow_headers: List[str]
    '''
    Configures the `Access-Control-Allow-Headers` header.

    Expects a list of allowed HTTP headers as strings, e.g. `["Content-Type", "Authorization"]`.
    '''
    allow_credentials: bool
    '''
    Configures the `Access-Control-Allow-Credentials` header.

    Set to `True` to pass the header, otherwise it is omitted.
    '''
    allow_origin_regex: str
    '''
    Same as `allow_origins`, but expects a string that will be compiled to a regular expression
    object in order to match the origin.
    '''
    expose_headers: List[str]
    '''
    Configures the `Access-Control-Expose-Headers` header.

    Expects a list of HTTP headers as strings. If not specified, no custom headers are exposed.
    '''
    max_age: int
    '''
    Configures the `Access-Control-Max-Age` header.

    Expects an integer. Defaults to `600` (10 min).
    '''


DEFAULT_CORS_OPTIONS: CorsOptions = {
    'allow_methods': [
        'GET',
    ],
    'allow_credentials': False,
    'max_age': 600,
}



import re
from typing import Union

from ...utils.colorize import c


class PestException(Exception):
    """Base class for all exceptions raised by Pest during setup

    This exception is not meant to be used outside of pest's internals, since it's not
    handled by any of the asgi exception handlers.
    """

    def __init__(
        self,
        *args: object,
        hint: Union[str, None] = None,
    ):
        if hint is not None:
            hint = f'{c("🐀 Hint ⇝ ", color="light_magenta")} {hint}'
            # code `text`
            hint = re.sub(r'`([^`]+)`', lambda m: c(m.group(1), color='green'), hint)
            # bold **text**
            hint = re.sub(r'\*\*([^*]+)\*\*', lambda m: c(m.group(1), attrs=['bold']), hint)
            # underline __text__
            hint = re.sub(r'__([^_]+)__', lambda m: c(m.group(1), attrs=['underline']), hint)

        super().__init__(*(*args, hint) if hint is not None else args)

    def __str__(self) -> str:
        return '\n'.join([str(arg) for arg in self.args])



"""
Module providing typed dicts for the metadata HandlerMeta:
- pest/metadata/types/handler_meta.py

ATTENTION:
This file was auto-generated by `task gen:types`.
Do not edit manually.
"""

from enum import Enum
from typing import Any, Dict, List, Sequence, Type, TypedDict, Union

from fastapi import Response


class HandlerMetaDict(TypedDict, total=False):
    response_model: Union[Any, None]
    '''🐀 ⇝ response model of the handler'''

    response_class: Union[Type[Response], None]
    '''🐀 ⇝ response class of the handler'''

    status_code: Union[int, None]
    '''🐀 ⇝ status code of the handler'''

    response_model_exclude_none: Union[bool, None]
    '''🐀 ⇝ exclude none values from the response model'''

    tags: Union[List[Union[str, Enum]], None]
    '''🐀 ⇝ tags of the handler'''

    dependencies: Union[Sequence[Any], None]
    '''🐀 ⇝ dependencies of the handler'''

    summary: Union[str, None]
    '''🐀 ⇝ summary of the handler'' '''

    description: Union[str, None]
    '''🐀 ⇝ description of the handler'''

    deprecated: Union[bool, None]
    '''🐀 ⇝ is the handler deprecated?'''

    name: Union[str, None]
    '''🐀 ⇝ name of the handler'''

    responses: Union[Dict[Union[int, str], Dict[str, Any]], None]
    '''🐀 ⇝ responses of the handler'''



from functools import wraps
from inspect import Parameter, getmembers, iscoroutinefunction, isfunction, signature
from typing import Any, Callable, Dict, List, Optional, Protocol, Tuple, Type, get_args

from fastapi import Request

from ..core.handler import HandlerFn
from ..exceptions.http.http import ForbiddenException
from ..metadata.meta import get_meta, get_meta_value
from ..metadata.types._meta import PestType

GuardCb = Callable[[Dict[str, Any]], None]


class Guard(Protocol):
    """🐀 ⇝ base guard protocol"""

    def can_activate(
        self, request: Request, *, context: Dict[str, Any], set_result: GuardCb
    ) -> bool:
        """🐀 ⇝ determines if the request can be activated by the current request"""
        ...


def use_guard(guard: Type[Guard]) -> Callable:
    """🐀 ⇝ decorator to apply a guard either to a single method or all methods in a class"""

    def decorator(target: Callable) -> Callable:
        if isinstance(target, type):  # If it's a class, apply to all methods
            return _apply_guard_to_class(target, guard)
        else:
            return _apply_guard_to_method(target, guard)

    return decorator


class GuardExtra(Dict[str, Any]):
    pass


def _extract_params(params: List[Parameter]) -> Tuple[Optional[Parameter], List[Parameter]]:
    """
    extracts the request and all parameters annotated with "guard_extra" from a list of parameters
    """
    request_param = None
    extra_params = []

    for param in params:
        if param.annotation == Request:
            request_param = param
        elif param.annotation is GuardExtra:
            extra_params.append(param)
        else:
            anns = get_args(param.annotation)
            if len(anns) == 0:
                continue

            typing, metas = anns[0], anns[1:]
            if typing == GuardExtra or GuardExtra in metas:
                extra_params.append(param)

    return request_param, extra_params


# applies the guard to a single method
def _apply_guard_to_method(func: Callable, guard: Type[Guard]) -> Callable:
    sig = signature(func)
    params: List[Parameter] = list(sig.parameters.values())

    # check if there's any parameter annotated with type Request
    request_parameter, extras = _extract_params(params)
    request_was_in_original_sig = request_parameter is not None

    if request_parameter is None:
        # add the parameter to the signature, annotated with type Request
        request_param_name = '__request__'
        params = [
            *params,
            Parameter(request_param_name, Parameter.POSITIONAL_OR_KEYWORD, annotation=Request),
        ]
    else:
        request_param_name = request_parameter.name

    # remove the extras the `params` list
    params = [param for param in params if param not in extras]

    @wraps(func)
    async def wrapper(*args: Any, **kwargs: Any) -> Any:
        meta = get_meta(func)

        # try to extract the request object from kwargs or args
        request = kwargs.get(request_param_name)
        if request is None:
            for arg in args:
                if isinstance(arg, Request):
                    request = arg
                    break

        if not request:
            raise ValueError('Request object not found in args or kwargs')

        extra_result: Dict[str, Any] = {}

        def set_result(result: Dict[str, Any]) -> None:
            nonlocal extra_result
            extra_result = result

        # apply the guard
        guard_instance = guard()
        if not guard_instance.can_activate(request, context=meta, set_result=set_result):
            raise ForbiddenException('Not authorized')

        # if the request was not in the original signature, remove it from args/kwargs
        if not request_was_in_original_sig:
            kwargs.pop(request_param_name, None)
            args = tuple([arg for arg in args if not isinstance(arg, Request)])

        # add the extra result to the signature
        for param in extras:
            if param.annotation is GuardExtra:
                kwargs[param.name] = extra_result
            else:
                kwargs[param.name] = extra_result.get(param.name, None)

        return await func(*args, **kwargs) if iscoroutinefunction(func) else func(*args, **kwargs)

    # update the signature to include the new 'request' parameter
    setattr(wrapper, '__signature__', sig.replace(parameters=params))
    return wrapper


# applies the guard to all methods in a class
def _apply_guard_to_class(cls: type, guard: Type[Guard]) -> type:
    members = getmembers(cls, lambda m: isfunction(m))
    handlers: List[HandlerFn] = []

    for _, method in members:
        meta_type = get_meta_value(method, key='meta_type', type=PestType, default=None)
        if meta_type == PestType.HANDLER:
            handlers.append(method)

    for handler in handlers:
        replacement = _apply_guard_to_method(handler, guard)
        setattr(cls, handler.__name__, replacement)

    return cls



try:
    from typing import Unpack, cast
except ImportError:
    from typing_extensions import Unpack, cast

from typing import Union

from ..core.application import PestApplication
from ..core.types.fastapi_params import FastAPIParams
from ..logging import LoggingOptions, log
from ..middleware.types import CorsOptions, MiddlewareDef
from ..utils.functions import getset
from . import post_app, pre_app
from .app_creator import make_app as make_app


class Pest:
    """🐀 ⇝ the main class of the framework, used to create and initialize a pest application"""

    @classmethod
    def create(
        cls,
        root_module: type,
        *,
        logging: Union[LoggingOptions, None] = None,
        middleware: MiddlewareDef = [],
        prefix: str = '',
        cors: Union[CorsOptions, None] = None,
        **fastapi_params: Unpack[FastAPIParams],
    ) -> PestApplication:
        """
        🐀 ⇝ creates and initializes a pest application
        #### Params
        - root_module: the root (entry point) module of the application
        - logging: logging options (needs `loguru` to be installed)
        - middleware: a list of middlewares to be applied to the application
        - prefix: the prefix for the application's routes
        """
        name = getset(cast(dict, fastapi_params), 'title', 'pest 🐀')

        pre_app.setup(logging=logging)
        log.info(f'Initializing {name}')

        app = make_app(fastapi_params, root_module, prefix=prefix, middleware=middleware)
        app = post_app.setup(app, cors=cors)

        log.debug(f'{name} initialized: \n{app}')
        return app



from typing import Dict, Iterable, Literal, Union

Color = Literal[
    'black',
    'grey',
    'red',
    'green',
    'yellow',
    'blue',
    'magenta',
    'cyan',
    'light_grey',
    'dark_grey',
    'light_red',
    'light_green',
    'light_yellow',
    'light_blue',
    'light_magenta',
    'light_cyan',
    'white',
]

Highlight = Literal[
    'on_black',
    'on_grey',
    'on_red',
    'on_green',
    'on_yellow',
    'on_blue',
    'on_magenta',
    'on_cyan',
    'on_light_grey',
    'on_dark_grey',
    'on_light_red',
    'on_light_green',
    'on_light_yellow',
    'on_light_blue',
    'on_light_magenta',
    'on_light_cyan',
    'on_white',
]

Attribute = Literal[
    'bold',
    'dark',
    'underline',
    'blink',
    'reverse',
    'concealed',
]

ATTRIBUTES: Dict[Attribute, int] = {
    'bold': 1,
    'dark': 2,
    'underline': 4,
    'blink': 5,
    'reverse': 7,
    'concealed': 8,
}

HIGHLIGHTS: Dict[Highlight, int] = {
    'on_black': 40,
    'on_grey': 40,
    'on_red': 41,
    'on_green': 42,
    'on_yellow': 43,
    'on_blue': 44,
    'on_magenta': 45,
    'on_cyan': 46,
    'on_light_grey': 47,
    'on_dark_grey': 100,
    'on_light_red': 101,
    'on_light_green': 102,
    'on_light_yellow': 103,
    'on_light_blue': 104,
    'on_light_magenta': 105,
    'on_light_cyan': 106,
    'on_white': 107,
}

COLORS: Dict[Color, int] = {
    'black': 30,
    'grey': 30,
    'red': 31,
    'green': 32,
    'yellow': 33,
    'blue': 34,
    'magenta': 35,
    'cyan': 36,
    'light_grey': 37,
    'dark_grey': 90,
    'light_red': 91,
    'light_green': 92,
    'light_yellow': 93,
    'light_blue': 94,
    'light_magenta': 95,
    'light_cyan': 96,
    'white': 97,
}

RESET = '\033[0m'


def c(
    text: object,
    color: Union[Color, None] = None,
    on_color: Union[Highlight, None] = None,
    attrs: Union[Iterable[Attribute], None] = None,
    no_color: bool = False,
) -> str:
    result = str(text)
    if no_color:
        return result

    fmt_str = '\033[%dm%s'
    if color is not None:
        result = fmt_str % (COLORS[color], result)

    if on_color is not None:
        result = fmt_str % (HIGHLIGHTS[on_color], result)

    if attrs is not None:
        for attr in attrs:
            result = fmt_str % (ATTRIBUTES[attr], result)

    result += RESET

    return result



from dataclasses import asdict
from inspect import Parameter, isclass, isfunction, signature
from typing import TYPE_CHECKING, Any, Callable, List, Tuple, Type, Union, get_args

try:
    from typing import TypeAlias
except ImportError:
    from typing_extensions import TypeAlias

from fastapi import Depends, Request
from fastapi.routing import APIRoute

from pest.di.injection import _Inject, inject

from ..exceptions.base.pest import PestException
from ..metadata.types.handler_meta import HandlerMeta
from ..middleware.di import scope_from
from ..utils.functions import clean_dict

if TYPE_CHECKING:  # pragma: no cover
    from ..metadata.types.module_meta import InjectionToken
    from .controller import Controller

HandlerFn: TypeAlias = Callable[..., Any]
HandlerTuple: TypeAlias = Tuple[HandlerFn, HandlerMeta]


def setup_handler(cls: Type['Controller'], handler: HandlerTuple) -> APIRoute:
    """
    Sets up a request handler.

    @internal
    """
    from ..decorators.dicts.handler_dict import HandlerMetaDict

    handler_fn, handler_meta = handler
    meta_dict = clean_dict(asdict(handler_meta), HandlerMetaDict)
    _patch_handler_fn(cls, handler_fn)

    route = APIRoute(
        endpoint=handler_fn,
        path=handler_meta.path,
        methods=handler_meta.methods,
        **meta_dict,
    )

    return route


def _patch_handler_fn(cls: Type['Controller'], handler: HandlerFn) -> None:
    """
    Changes the signature of a route's endpoint to ensure that FastAPI
    performs dependency injection correctly and doesn't expect a `self`
    parameter as a query parameter. To do so, we replace the first
    parameter of the endpoint with a `Depends` object that resolves the
    controller instance using `rodi`.

    It also replaces any parameter annotated with `inject` or with a
    default value of `_Inject` with a `Depends` object that resolves the
    value using the `module`'s rodi container. This last step is
    performed by `PestFastAPIInjector`.

    @internal
    """

    controller = PestFastAPIInjector(token=cls, controller=cls)

    old_endpoint = handler
    old_signature = signature(old_endpoint)
    old_parameters: List[Parameter] = list(old_signature.parameters.values())
    old_first_parameter = old_parameters[0]
    new_first_parameter = old_first_parameter.replace(default=Depends(controller))
    new_parameters = [_get_new_param(cls, parameter) for parameter in old_parameters[1:]]
    new_signature = old_signature.replace(parameters=[new_first_parameter] + new_parameters)

    setattr(handler, '__signature__', new_signature)


def _get_new_param(ctrl: type, parameter: Parameter) -> Parameter:
    """
    If the parameter is expected to be injected by `pest` (either by annotating it with `inject` or
    by having a default value of `_Inject` returned by `= inject(something)`), we replace it with a
    FastAPI's `Depends` that resolves the value using the `module`'s rodi container. This last step
    is performed by `PestFastAPIInjector`.

    @internal
    """
    if parameter.annotation is not Parameter.empty or isinstance(parameter.default, _Inject):
        pest_anns = _get_pest_injection(parameter)
        if pest_anns is not None:
            # it has a pest injection annotation

            if len(pest_anns) > 1:
                raise PestException(
                    'Multiple injection annotations are not allowed!',
                    hint=f'Parameter {parameter.name} has {len(pest_anns)} injection annotation!',
                )

            annotation = pest_anns[0]
            if annotation.token is not None:
                # if the token is a function, we replace it with a FastAPI's `Depends(dep)` call
                if isfunction(annotation.token):
                    parameter = parameter.replace(default=Depends(annotation.token))
                elif isclass(annotation.token):
                    # otherwise we replace the parameter with a `Depends` on `PestFastAPIInjector`
                    # which will try to resolve the value from the `module`'s container
                    parameter = parameter.replace(
                        default=Depends(
                            PestFastAPIInjector(controller=ctrl, token=annotation.token)
                        )
                    )
                else:
                    raise PestException(
                        'Invalid injection annotation token!',
                        hint=f'Parameter {parameter.name} has an invalid injection token!',
                    )
    return parameter.replace(kind=Parameter.KEYWORD_ONLY)


def _get_pest_injection(parameter: Parameter) -> Union[List[_Inject], None]:
    """
    checks if the parameter is annotated with `inject` or has a default value of `_Inject`.
    If so, it returns a list of `_Inject` annotations, otherwise it returns `None`.

    @internal
    """

    if isinstance(parameter.default, _Inject):
        token = (
            parameter.default.token
            if parameter.default.token is not None
            else parameter.annotation
            if parameter.annotation is not Parameter.empty
            else None
        )
        annotations = (token, inject)
    else:
        annotations = get_args(parameter.annotation)

    if len(annotations) == 0:
        return None

    token, anns = annotations[0], annotations[1:]

    if token is None:
        raise PestException(
            'Missing injection annotation token!',
            hint=f'Parameter {parameter.name} has no injection annotation!',
        )

    anns = [
        annotation if isinstance(annotation, _Inject) else _Inject(token)
        for annotation in anns
        if isinstance(annotation, _Inject) or annotation is inject
    ]

    for annotation in anns:
        if annotation.token is None:
            annotation.token = token

    return anns if len(anns) > 0 else None


class PestFastAPIInjector:
    """
    Injector to be used with fastapi's `Depends`

    This injector is used to bypass the `Depends` mechanism of FastAPI
    and inject stuff into the handler's signature ourselves using
    the `module`'s container.

    @internal
    """

    def __init__(self, controller: Type['Controller'], token: 'InjectionToken'):
        self.controller = controller
        self.token = token

    async def __call__(self, request: Request) -> Any:
        """🐀 ⇝ returns the `controller` to be injected 💉"""
        from .controller import module_of

        scope = scope_from(request)
        module = module_of(self.controller)
        di_result = await module.aget(self.token, scope)

        return di_result



from typing import Union

from dij import ActivationScope
from fastapi import Request, Response
from starlette.middleware.base import RequestResponseEndpoint

SCOPE_KEY = '__di_scope__'
'''🐀 ⇝ the key used to store the di activation scope in a request'''


def scope_from(request: Request) -> Union[ActivationScope, None]:
    """obtains the di activation scope from a request"""
    return getattr(request.state, SCOPE_KEY, None)


async def di_scope_middleware(request: Request, call_next: RequestResponseEndpoint) -> Response:
    """
    middleware that injects a di activation scope into the request. Allows the
    injection of request-scoped dependencies into controllers, services and other
    middlewares.
    """

    with ActivationScope() as scope:
        if scope.scoped_services is not None:
            setattr(request.state, SCOPE_KEY, scope)

        return await call_next(request)



from typing import Union

from fastapi.encoders import jsonable_encoder
from fastapi.exceptions import RequestValidationError, WebSocketRequestValidationError
from fastapi.utils import is_body_allowed_for_status_code
from fastapi.websockets import WebSocket
from pydantic import ValidationError
from starlette.exceptions import HTTPException
from starlette.requests import Request
from starlette.responses import JSONResponse, Response
from starlette.status import WS_1008_POLICY_VIOLATION

from ..utils.functions import dump_model
from .http.http import ExceptionResponse, PestHTTPException
from .http.status import HTTPStatusEnum, http_status


async def http(request: Request, exc: HTTPException) -> Response:
    """handles both `HTTPException` (fastapi) and `PestHTTPException` (pest)"""

    headers = getattr(exc, 'headers', None)
    if not is_body_allowed_for_status_code(exc.status_code):
        return Response(status_code=exc.status_code, headers=headers)

    if isinstance(exc, PestHTTPException):
        content = vars(exc)
    else:
        content = vars(
            PestHTTPException(
                status_code=exc.status_code,
                detail=exc.detail,
                headers=headers,
            )
        )

    return JSONResponse(
        status_code=exc.status_code,
        content=content,
        headers=headers,
    )


async def request_validation(
    request: Request, exc: Union[ValidationError, RequestValidationError]
) -> JSONResponse:
    """handles both `RequestValidationError` (fastapi) and `ValidationError` (pydantic)"""

    stat = http_status(HTTPStatusEnum.HTTP_400_BAD_REQUEST)

    messages = [
        f"{', '.join([str(elem) for elem in err['loc']])}: {err['msg'].capitalize()}"
        for err in exc.errors()
    ]

    return JSONResponse(
        status_code=stat.code,
        content=vars(ExceptionResponse(code=stat.code, error=stat.phrase, message=messages)),
    )


async def websocket_request_validation(
    websocket: WebSocket, exc: WebSocketRequestValidationError
) -> None:
    await websocket.close(code=WS_1008_POLICY_VIOLATION, reason=jsonable_encoder(exc.errors()))


async def the_rest(request: Request, exc: Exception) -> Response:
    """handles all other exceptions

    returns a generic 500 response and logs the exception message, just in case it
    was an
    """

    stat = http_status(HTTPStatusEnum.HTTP_500_INTERNAL_SERVER_ERROR)

    return JSONResponse(
        status_code=stat.code,
        content=dump_model(ExceptionResponse(code=stat.code, error=stat.phrase)),
    )



from .decorators.controller import api, controller, ctrl, router, rtr
from .decorators.guard import Guard, GuardCb, GuardExtra, use_guard
from .decorators.handler import delete, get, head, options, patch, post, put, trace
from .decorators.module import dom, domain, mod, module
from .factory import Pest
from .metadata.types.injectable_meta import (
    ClassProvider,
    ExistingProvider,
    FactoryProvider,
    ProviderBase,
    Scope,
    SingletonProvider,
    ValueProvider,
)
from .utils.decorators import meta

guard = use_guard

__all__ = [
    'Pest',
    # decorators - module
    'module',
    'mod',
    'domain',
    'dom',
    # decorators - handler
    'get',
    'post',
    'put',
    'delete',
    'patch',
    'options',
    'head',
    'trace',
    # decorators - controller
    'controller',
    'ctrl',
    'router',
    'rtr',
    'api',
    # decorators - utils
    'meta',
    # decorators - guard
    'Guard',
    'GuardCb',
    'GuardExtra',
    'use_guard',
    'guard',
    # meta - providers
    'ProviderBase',
    'ClassProvider',
    'ValueProvider',
    'SingletonProvider',
    'FactoryProvider',
    'ExistingProvider',
    'Scope',
]



"""
Module providing typed dicts for the metadata ControllerMeta:
- pest/metadata/types/controller_meta.py

ATTENTION:
This file was auto-generated by `task gen:types`.
Do not edit manually.
"""

from enum import Enum
from typing import Any, Callable, List, Sequence, TypedDict, Union


class ControllerMetaDict(TypedDict, total=False):
    tags: Union[List[Union[str, Enum]], None]
    '''🐀 ⇝ tags of the controller'''

    redirect_slashes: Union[bool, None]
    '''🐀 ⇝ redirect slashes?'''

    on_startup: Union[Sequence[Callable[[], Any]], None]
    '''🐀 ⇝ on startup events'''

    on_shutdown: Union[Sequence[Callable[[], Any]], None]
    '''🐀 ⇝ lifespan of the controller'''

    deprecated: Union[bool, None]
    '''🐀 ⇝ is the controller deprecated?'''

    include_in_schema: Union[bool, None]
    '''🐀 ⇝ include in schema?'''



"""
Module providing typed dicts for the metadata ModuleMeta:
- pest/metadata/types/module_meta.py

ATTENTION:
This file was auto-generated by `task gen:types`.
Do not edit manually.
"""

from typing import List, Type, TypedDict, Union

from ...core.controller import Controller
from ...metadata.types.module_meta import InjectionToken, Provider


class ModuleMetaDict(TypedDict, total=False):
    imports: Union[List[type], None]
    '''🐀 ⇝ list of modules to be imported'''

    providers: Union[List[Provider], None]
    '''🐀 ⇝ list of providers to be registered'''

    exports: Union[List[InjectionToken], None]
    '''🐀 ⇝ list of providers to be exported'''

    controllers: Union[List[Type[Controller]], None]
    '''🐀 ⇝ list of controllers to be registered'''



from typing import Any, Callable, Dict, List, Type, TypeVar, Union, cast

from pydantic import BaseModel

T = TypeVar('T', bound=Dict)


def denone(x: T) -> T:
    """remove all keys with `None` values from a dictionary"""
    return cast(T, {k: v for k, v in x.items() if v is not None})


def clean_dict(d: Dict[str, Any], type: Type) -> Dict[str, Any]:
    """
    performs a cleaning operation on a dictionary to:
    1. remove all keys with `None` values
    2. keep only keys that are in the `TypedDict` type
    """
    return denone({k: v for k, v in d.items() if k in type.__annotations__})


def keep_keys(d: Dict[str, Any], keys: List[str]) -> Dict[str, Any]:
    """keep only the specified keys in a dictionary"""
    return denone({k: v for k, v in d.items() if k in keys})


def drop_keys(d: Dict[str, Any], keys: List[str]) -> Dict[str, Any]:
    """drop the specified keys from a dictionary"""
    return denone({k: v for k, v in d.items() if k not in keys})


def getset(d: dict, key: str, default: Any) -> Any:
    """get a value from a dictionary or set it if it does not exist, then return it"""
    if key not in d:
        d[key] = default
    return d[key]


def set_if_none(d: dict, key: str, value: Any) -> None:
    """set a value in a dictionary only if it does not exist on it already"""
    if key not in d:
        d[key] = value


def dump_model(model: BaseModel) -> Dict[str, Any]:
    """dump a pydantic model to a dict

    HACK: this is a temporary function to support the new pydantic's model_dump() function
    while not breaking compatibility with pydantic@<1. This function works as a centralised
    way of future changes once pydantic<2 is dropped and model_dump() is the only
    function available. When that happens, this function will be removed and all calls to it
    will be replaced by `model.model_dump()`.
    """

    return (
        model.model_dump(exclude_none=True, exclude_unset=True)
        if hasattr(model, 'model_dump')
        else model.dict(exclude_none=True, exclude_unset=True)
    )


def model_schema(model: Type[BaseModel]) -> Dict[str, Any]:
    """return a pydantic model's schema

    HACK: this is a temporary function to support the new pydantic's model_json_schema()
    function while not breaking compatibility with pydantic@<2. This function works as a
    centralised way of future changes once pydantic<2 is dropped and model_json_schema() is
    the only function available. When that happens, this function will be removed and all
    calls to it will be replaced by `model.model_json_schema()`.
    """

    return model.model_json_schema() if hasattr(model, 'model_json_schema') else model.schema()


class classproperty:
    """basic implementation of a class property decorator"""

    def __init__(self, method: Callable) -> None:
        self.method = method

    def __get__(self, obj: Any, cls: Union[Type, None] = None) -> ...:
        if cls is None:
            cls = type(obj)
        return self.method(cls)



from .base.pest import PestException
from .http.http import (
    BadGatewayException,
    BadRequestException,
    ConflictException,
    ExceptionResponse,
    ForbiddenException,
    GatewayTimeoutException,
    GoneException,
    HttpVersionNotSupportedException,
    ImATeapotException,
    InternalServerErrorException,
    MethodNotAllowedException,
    NotAcceptableException,
    NotFoundException,
    NotImplementedException,
    PayloadTooLargeException,
    PestHTTPException,
    PreconditionFailedException,
    ProxyAuthenticationRequiredException,
    RequestTimeoutException,
    ServiceUnavailableException,
    UnauthorizedException,
    UnprocessableEntityException,
    UnsupportedMediaTypeException,
    exc_response,
)

__all__ = [
    'ExceptionResponse',
    'PestHTTPException',
    'BadRequestException',
    'UnauthorizedException',
    'ForbiddenException',
    'NotFoundException',
    'MethodNotAllowedException',
    'NotAcceptableException',
    'ProxyAuthenticationRequiredException',
    'RequestTimeoutException',
    'ConflictException',
    'GoneException',
    'PreconditionFailedException',
    'PayloadTooLargeException',
    'UnsupportedMediaTypeException',
    'ImATeapotException',
    'UnprocessableEntityException',
    'InternalServerErrorException',
    'NotImplementedException',
    'BadGatewayException',
    'ServiceUnavailableException',
    'GatewayTimeoutException',
    'HttpVersionNotSupportedException',
    'PestException',
    'exc_response',
]



from typing import Any, Callable, Mapping, Type, TypeVar, Union

from dacite import Config, from_dict

from ..metadata.meta import inject_metadata
from ..metadata.types._meta import Meta

T = TypeVar('T')
K = TypeVar('K')
Cls = TypeVar('Cls', bound=Type[Any])


def singleton(cls: Type[T]) -> Type[T]:
    """
    makes the decorated class a singleton, ensuring
    that only one instance of the class is ever created.

    ```
    @singleton
    class MyClass: pass

    foo = MyClass()
    bar = MyClass()
    print(foo is bar)  # True
    ```
    """

    new = cls.__new__

    def constructor(cls: Any, *args: Any) -> T:
        instance = getattr(cls, '__instance__', None)
        if instance is None:
            instance = new(cls)
            setattr(cls, '__instance__', instance)
        return instance

    setattr(cls, '__new__', constructor)
    return cls


make_singleton = singleton


def _inject_class(base: Type[T]) -> Callable[..., Type[T]]:
    return lambda cls: type(cls.__name__, (base, cls) + cls.__bases__, dict(cls.__dict__))


def mixin(base: Type[T]) -> Callable[..., Type[T]]:
    """
    Utility decorator to inject a class as a mixin and combine it with the
    decorated class.

    The difference between this and `use_base` is that this
    decorator will cast the decorated class to the type of the base class.
    """
    return _inject_class(base=base)


def use_base(base: type) -> Callable[[T], Callable[..., T]]:
    """
    Utility decorator to inject a class as a mixin and combine it with the
    decorated class.

    The difference between this and `mixin` is that this
    decorator will not cast the decorated class to the type of the base class.
    """
    return _inject_class(base=base)


def meta_decorator_mixin(
    meta_type: Type[Meta],
    meta: Mapping[str, Any],
    base: type,
    singleton: bool = False,
) -> Callable[[Cls], Type[Cls]]:
    """
    makes a decorator that sets the meta of a class
    and optionally makes it inherit from a base class
    """

    def wrapper(cls: Cls) -> Type[Cls]:
        if singleton:
            cls = make_singleton(cls)

        inject_metadata(cls, from_dict(meta_type, meta, config=Config(check_types=False)))
        return mixin(base)(cls)

    return wrapper


def meta_decorator(
    meta_type: Type[Meta],
    meta: Mapping[str, Any],
    base: Union[type, None] = None,
    singleton: bool = False,
) -> Callable:
    """
    makes a decorator that sets the meta of a class
    and optionally makes it inherit from a base class
    """

    def wrapper(cls: Cls) -> Callable[..., Cls]:
        if singleton:
            cls = make_singleton(cls)

        inject_metadata(cls, from_dict(meta_type, meta, config=Config(check_types=False)))
        return use_base(base)(cls) if base is not None else cls

    return wrapper


def compose_decorators(*decs: Callable[..., Any]) -> Callable[..., Any]:
    def deco(f: Callable[..., Any]) -> Callable[..., Any]:
        for dec in reversed(decs):
            f = dec(f)
        return f

    return deco



from dataclasses import asdict
from typing import Any, Callable, Mapping, Type, TypeVar, Union

from ..metadata.meta import inject_metadata
from .protocols import DataclassInstance

Cls = TypeVar('Cls', bound=Type[Any])
Fn = TypeVar('Fn', bound=Callable[..., Any])


def meta(
    meta: Union[Mapping[str, Any], DataclassInstance]
) -> Callable[[Union[Cls, Fn]], Union[Cls, Fn]]:
    """🐀 » injects metadata into a class"""

    def decorator(callable: Union[Cls, Fn]) -> Union[Cls, Fn]:
        if isinstance(meta, Mapping):
            metadata = meta
        else:
            metadata = asdict(meta)

        inject_metadata(callable, metadata)
        return callable

    return decorator



from __future__ import annotations

import logging as pylogging
import sys
from fnmatch import fnmatch
from pprint import pformat
from typing import Callable, List, Tuple, Union

try:
    from typing import Unpack, cast
except ImportError:
    from typing_extensions import Callable, Unpack, cast

from ...exceptions.base.pest import PestException
from ...utils.functions import set_if_none
from .. import LoggingOptions, LogLevel, SinkOptions

try:
    import loguru
    from loguru import logger as loguru_logger
    from loguru._defaults import env
except ImportError as e:  # noqa: F841
    raise PestException(
        'Failed to import loguru',
        hint=(
            'Install `loguru` to use logging configuration '
            'system: `pip install pest[loguru]`, `poetry add pest[loguru]` '
            'or just `pip install loguru` or `poetry add loguru`'
        ),
    )

FORMAT = env(
    'LOGURU_FORMAT',
    str,
    '<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | '
    '<level>{level: <8}</level> | '
    '<cyan>{name}</cyan>:<cyan>{line}</cyan> - <level>{message}</level>'
    '{exception}',
)


class _InterceptorHandler(pylogging.Handler):
    def get_exc(self, record: pylogging.LogRecord) -> Tuple[str, BaseException] | None:
        exc = record.exc_info
        if (
            exc is None
            or (exc is not None and exc[0] is None)
            or (exc is not None and exc[1] is None)
        ):
            return

        return (exc[0].__name__, exc[1])

    def emit(self, record: pylogging.LogRecord) -> None:
        """intercepts a python log (logging lib) and sends it to loguru"""
        try:
            level = loguru_logger.level(record.levelname).name
        except ValueError:
            level = record.levelno

        frame, depth = pylogging.currentframe(), 0
        while frame and (depth == 0 or frame.f_code.co_filename == pylogging.__file__):
            frame = frame.f_back
            depth += 1

        log = loguru_logger.bind(access=True if record.name == 'uvicorn.access' else None)

        e = self.get_exc(record)

        log.opt(
            depth=depth,
            exception=record.exc_info,
        ).log(level, f'{e[0]}: {e[1]}\n\n' if e else record.getMessage())


def no_access(record: loguru.Record) -> bool:
    return record['extra'].get('access') is None


def format_record(record: loguru.Record) -> str:
    format_string: str = cast(str, FORMAT)
    if record['extra'].get('payload') is not None:
        record['extra']['payload'] = pformat(
            record['extra']['payload'], indent=4, compact=True, width=88
        )
        format_string += '\n<level>{extra[payload]}</level>'
    return format_string + '\n'


class Loguru:
    """🐀 ⇝ loguru logging system"""

    @classmethod
    def setup(cls, **options: Unpack[LoggingOptions]) -> None:
        """🐀 ⇝ sets up `loguru`. Needs `loguru` library to be installed"""
        global FORMAT
        intercept = options.get('intercept', [])
        shush = options.get('shush', [])
        level = options.get('level', LogLevel.INFO)
        fmt = options.get('format', None)
        access_log = options.get('access_log', False)
        verbose = options.get('verbose', False)
        sinks = options.get('sinks', [])

        # if a format string is provided, we replace the default
        if type(fmt) is str:
            FORMAT = fmt

        cls.__config_standard_interception(intercept, verbose)
        cls.__configure_shush(shush)
        cls.__config_sinks(sinks, level, access_log, fmt)

    @classmethod
    def __config_standard_interception(
        cls, intercept: List[str | Tuple[str, LogLevel]], verbose: bool
    ) -> None:
        """configures interception of python loggers towards `loguru`"""
        interceptor = _InterceptorHandler()
        intercepted_loggers: List[Tuple[pylogging.Logger, int]] = []
        pylogging.basicConfig(handlers=[pylogging.NullHandler()])

        # clear handlers for all loggers in standard library
        for name in pylogging.root.manager.loggerDict:
            pylogging.getLogger(name).handlers.clear()

        if verbose:
            # if verbose, intercept all loggers and set them to DEBUG
            for name in pylogging.root.manager.loggerDict:
                logger = pylogging.getLogger(name)
                if '.' not in name:
                    logger.handlers = [interceptor]
                logger.propagate = True
                logger.setLevel(pylogging.DEBUG)
                logger.disabled = False
        else:
            for wc in intercept:
                lvl = None
                if isinstance(wc, tuple):
                    wc, lvl = wc

                # make a list of all loggers that match the wildcard
                to_intercept = [
                    (pylogging.getLogger(name), lvl or pylogging.getLogger(name).level)
                    for name in pylogging.root.manager.loggerDict
                    if fnmatch(name, wc)
                ]

                if len(to_intercept) == 0:
                    to_intercept.append((pylogging.getLogger(wc), pylogging.getLogger(wc).level))

                intercepted_loggers.extend(to_intercept)

        # intercept all loggers in the list
        for logger, lvl in intercepted_loggers:
            logger.handlers = [interceptor]
            logger.propagate = False
            logger.disabled = False
            logger.setLevel(lvl)

    @classmethod
    def __configure_shush(cls, shush: List[str]) -> None:
        """disables loggers by name"""
        for logger_name in shush:
            pylogging.getLogger(logger_name).disabled = True
            loguru_logger.disable(logger_name)

    @classmethod
    def __config_sinks(
        cls,
        sinks: List[SinkOptions],
        level: LogLevel,
        access_log: bool,
        fmt: Union[str, Callable[[loguru.Record], str], None],
    ) -> None:
        """configures loguru sinks (handlers)"""
        fmt_func = fmt if callable(fmt) else format_record

        loguru_logger.configure(handlers=[])
        loguru_logger.add(
            sys.stdout,
            colorize=True,
            diagnose=False,
            format=fmt_func,
            level=level,
            filter=lambda record: ((True if access_log else no_access(record))),
        )

        for sink in sinks:
            set_if_none(cast(dict, sink), 'format', fmt_func)
            loguru_logger.add(**sink)  # type: ignore



from typing import Tuple, Union

from .colorize import c


def as_tree(
    obj: Union[object, Tuple[str, str]],
    prefix: str = '',
    is_last: bool = True,
    with_providers: bool = True,
    with_controllers: bool = True,
) -> str:
    """returns a tree representation of the module and its submodules"""
    if isinstance(obj, tuple):
        t, n = obj
        if t == 'provider':
            obj_name = '│' + c(f'{" ○ " if prefix else ""}{n}', color='magenta')
        else:
            obj_name = '│' + c(f'{" □ " if prefix else ""}{n}', color='blue')
    else:
        obj_name = f'{"├─ " if prefix else ""}{obj.__class__.__name__}'

    result = ''

    if prefix:
        result += prefix + obj_name + '\n'
    else:
        result += f'{c(obj_name, color="green", attrs=["underline"])} 🐀' + '\n' + '    │\n'

    prefix += '│   ' if not is_last else '    '

    if with_providers:
        providers = getattr(obj, 'providers', [])
        for i, provider in enumerate(providers):
            provider_name = _get_provider_name(provider)
            result += as_tree(('provider', provider_name), prefix, i == len(providers) - 1)

    if with_controllers:
        controllers = getattr(obj, 'controllers', [])
        for i, controller in enumerate(controllers):
            controller_name = controller.__name__
            result += as_tree(('controller', controller_name), prefix, i == len(controllers) - 1)

    imports = getattr(obj, 'imports', [])
    for i, sub_module in enumerate(imports):
        result += as_tree(sub_module, prefix, i == len(imports) - 1)

    return result


def _get_provider_name(provider: type) -> str:
    injection_token = getattr(provider, 'provide', None)

    if injection_token is None:
        return provider.__name__

    if isinstance(injection_token, str):
        return injection_token

    return injection_token.__name__



from typing import Any, Dict, List, Optional, Tuple, Type, TypeVar, Union, cast

from dij import ActivationScope, Container, ServiceLifeStyle
from fastapi import APIRouter

import pest.utils.module as module_utils
from pest.metadata.types._meta import PestType

from ..exceptions.base.pest import PestException
from ..metadata.meta import get_meta
from ..metadata.types.injectable_meta import (
    ClassProvider,
    ExistingProvider,
    FactoryProvider,
    ValueProvider,
)
from ..metadata.types.module_meta import InjectionToken, ModuleMeta, Provider
from ..utils.functions import classproperty
from .common import PestPrimitive
from .controller import Controller, router_of, setup_controller
from .types.status import Status


def parent_of(module: 'Module') -> Optional['Module']:
    """returns the parent module of a given module"""
    if not isinstance(module, Module):
        raise PestException(
            f'{module.__name__} is not a module.',
            hint=f'decorate `{module.__name__}` with the `@module` decorator '
            '(or one of its aliases)',
        )

    return module.__parent_module__


def contained_in(module: 'Module') -> List[Tuple[InjectionToken, Any]]:
    """returns the providers contained in a given module"""
    if not isinstance(module, Module):
        raise PestException(
            f'{module.__name__} is not a module.',
            hint=f'decorate `{module.__name__}` with the `@module` decorator '
            '(or one of its aliases)',
        )

    return list(module.container)


def setup_module(
    clazz: type,
    parent_clazz: Optional['Module'] = None,
) -> 'Module':
    """
    functions that sets up a module. Avoids accessing the
    module's `__setup_module__` method directly
    """
    if not issubclass(clazz, Module):
        raise PestException(
            f'{clazz.__name__} is not a module.',
            hint=(
                f'decorate `{clazz.__name__}` with the `@module` decorator (or one of its aliases)'
            ),
        )

    if parent_clazz is not None and not isinstance(parent_clazz, Module):
        raise PestException(
            f'{parent_clazz.__name__} is not a module.',
            hint=(
                f'decorate `{parent_clazz.__name__}` with the `@module` '
                'decorator (or one of its aliases)'
            ),
        )

    module = clazz()
    module.__setup_module__(parent_clazz)
    return module


T = TypeVar('T')


class Module(PestPrimitive):
    __imported__providers__: Dict[InjectionToken, 'Module']
    __parent_module__: Optional['Module']
    imports: List['Module']
    container: Container
    providers: List[Any]
    exports: List[InjectionToken]
    controllers: List[Type[Controller]]

    @property
    def routers(self) -> List[APIRouter]:
        return self.__get_routers()

    @classproperty
    def __pest_object_type__(cls) -> PestType:
        return PestType.MODULE

    def __init__(self) -> None:
        self.__class_status__ = Status.NOT_SETUP
        self.__imported__providers__ = {}
        self.imports = []
        self.providers = []
        self.exports = []
        self.container = Container(strict=False)
        self.controllers = []

    def __setup_module__(self, parent: Optional['Module']) -> None:
        if self.__class_status__ != Status.NOT_SETUP:
            return
        self.__class_status__ = Status.SETTING_UP

        if parent is not None and isinstance(parent, Module):
            self.__parent_module__ = parent

        # get module metadata
        meta: ModuleMeta = get_meta(self.__class__, ModuleMeta)

        # set internal properties
        self.providers = meta.providers if meta.providers else []
        self.exports = meta.exports if meta.exports else []
        self.controllers = meta.controllers if meta.controllers else []

        # register providers in the di container
        for provider in self.providers:
            self.register(provider)

        # register controllers in the di container
        for controller in self.controllers:
            setup_controller(controller, self)
            self.register(controller)

        # register parent providers as factory providers here, so that we
        # can have access to services provided by the parent module
        # or globally (in the root module)
        if parent is not None:

            def create_factory(provider: InjectionToken) -> Any:
                async def resolve_from_parent() -> Any:
                    resolved = await parent.aget(cast(InjectionToken, provider))
                    return resolved

                return resolve_from_parent

            for provider, _ in contained_in(parent):
                self.register(
                    FactoryProvider(provide=provider, use_factory=create_factory(provider))
                )

        # setup child modules
        for child in meta.imports if meta.imports else []:
            child_instance = setup_module(child, self)
            self.imports += [child_instance]

        # register providers exported by child modules
        for imported_module in self.imports:
            for exported_provider in imported_module.exports:
                self.__imported__providers__[exported_provider] = imported_module

        # we're done
        self.__class_status__ = Status.READY

    def register(self, provider: Provider) -> None:
        if isinstance(provider, ClassProvider):
            self.container.bind_types(
                provider.provide,
                provider.use_class,
                life_style=(
                    provider.scope if provider.scope is not None else ServiceLifeStyle.TRANSIENT
                ),
            )
        elif isinstance(provider, FactoryProvider):
            self.container.register_factory(
                factory=provider.use_factory,
                return_type=provider.provide,
                life_style=(
                    provider.scope if provider.scope is not None else ServiceLifeStyle.TRANSIENT
                ),
            )
        elif isinstance(provider, ValueProvider):
            self.container.add_instance(
                declared_class=provider.provide,
                instance=provider.use_value,
            )
        elif isinstance(provider, ExistingProvider):
            self.container.add_alias(name=provider.provide, desired_type=provider.use_existing)
        else:
            self.container.add_transient(provider)

    def can_provide(self, token: InjectionToken) -> bool:
        if token in self.__imported__providers__ or token in self.container:
            return True
        return False

    def get(
        self, token: InjectionToken[T], scope: Union[ActivationScope, None] = None, **kwargs: Any
    ) -> T:
        if token in self.__imported__providers__:
            return self.__get_from_imported(token, scope=scope)

        return self.container.resolve(token, scope=scope, **kwargs)

    async def aget(self, token: InjectionToken[T], scope: Union[ActivationScope, None] = None) -> T:
        if token in self.__imported__providers__:
            return await self.__aget_from_imported(token, scope=scope)

        return await self.container.aresolve(token, scope=scope)

    def __get_from_imported(
        self, token: InjectionToken[T], scope: Union[ActivationScope, None] = None
    ) -> T:
        return self.__imported__providers__[token].get(token, scope=scope)

    async def __aget_from_imported(
        self, token: InjectionToken[T], scope: Union[ActivationScope, None] = None
    ) -> T:
        return await self.__imported__providers__[token].aget(token, scope=scope)

    def __get_routers(self) -> List[APIRouter]:
        routers = []
        for controller in self.controllers:
            routers += [router_of(controller)]

        for child in self.imports:
            routers += child.__get_routers()

        return routers

    def __str__(self) -> str:
        return module_utils.as_tree(self)



from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Dict, List, Sequence, Type, Union

from fastapi import Response

from ._meta import Meta, PestType


class HttpMethod(str, Enum):
    GET = 'GET'
    HEAD = 'HEAD'
    POST = 'POST'
    PUT = 'PUT'
    DELETE = 'DELETE'
    CONNECT = 'CONNECT'
    OPTIONS = 'OPTIONS'
    TRACE = 'TRACE'
    PATCH = 'PATCH'


@dataclass
class HandlerMeta(Meta):
    meta_type: PestType = field(default=PestType.HANDLER, init=False, metadata={'expose': False})
    methods: List[str] = field(metadata={'expose': False})
    '''🐀 ⇝ HTTP methods of the handler'''
    path: str = field(metadata={'expose': False})
    '''🐀 ⇝ path of the handler'''
    response_model: Union[Any, None]
    '''🐀 ⇝ response model of the handler'''
    response_class: Union[Type[Response], None]
    '''🐀 ⇝ response class of the handler'''
    status_code: Union[int, None]
    '''🐀 ⇝ status code of the handler'''
    response_model_exclude_none: Union[bool, None]
    '''🐀 ⇝ exclude none values from the response model'''
    tags: Union[List[Union[str, Enum]], None]
    '''🐀 ⇝ tags of the handler'''
    dependencies: Union[Sequence[Any], None]
    '''🐀 ⇝ dependencies of the handler'''
    summary: Union[str, None]
    '''🐀 ⇝ summary of the handler'' '''
    description: Union[str, None]
    '''🐀 ⇝ description of the handler'''
    deprecated: Union[bool, None]
    '''🐀 ⇝ is the handler deprecated?'''
    name: Union[str, None]
    '''🐀 ⇝ name of the handler'''
    responses: Union[Dict[Union[int, str], Dict[str, Any]], None]
    '''🐀 ⇝ responses of the handler'''



from enum import Enum
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    Set,
    Type,
    Union,
)

from fastapi import APIRouter, params
from fastapi.datastructures import Default, DefaultPlaceholder
from fastapi.routing import APIRoute
from fastapi.types import IncEx
from fastapi.utils import (
    generate_unique_id,
)
from starlette.responses import JSONResponse, Response
from starlette.routing import BaseRoute

if TYPE_CHECKING:  # pragma: no cover
    from ...core.controller import Controller


class PestRouter(APIRouter):
    """
    Extends the `APIRouter` class from FastAPI to handle / at the end of API routes.
    By default, FastAPI redirects routes that end in / to the route without /. This
    class avoids that behavior and for each route that is added, it adds an alternative
    route with or without /, as appropriate.
    """

    routes: List[APIRoute]
    controller: Type['Controller']

    def add_api_route(
        self,
        path: str,
        endpoint: Callable[..., Any],
        *,
        response_model: Any = Default(None),
        status_code: Optional[int] = None,
        tags: Optional[List[Union[str, Enum]]] = None,
        dependencies: Optional[Sequence[params.Depends]] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        response_description: str = 'Successful Response',
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        deprecated: Optional[bool] = None,
        methods: Optional[Union[Set[str], List[str]]] = None,
        operation_id: Optional[str] = None,
        response_model_include: Optional[IncEx] = None,
        response_model_exclude: Optional[IncEx] = None,
        response_model_by_alias: bool = True,
        response_model_exclude_unset: bool = False,
        response_model_exclude_defaults: bool = False,
        response_model_exclude_none: bool = False,
        include_in_schema: bool = True,
        response_class: Union[Type[Response], DefaultPlaceholder] = Default(JSONResponse),
        name: Optional[str] = None,
        route_class_override: Optional[Type[APIRoute]] = None,
        callbacks: Optional[List[BaseRoute]] = None,
        openapi_extra: Optional[Dict[str, Any]] = None,
        generate_unique_id_function: Union[Callable[[APIRoute], str], DefaultPlaceholder] = Default(
            generate_unique_id
        ),
        **kwargs: Any,
    ) -> None:
        """
        Registra un endpoint de la API con la ruta proporcionada y con su ruta alternativa
        dependiendo de si la ruta proporcionada termina en `/` o no.

        Por ejemplo, si se registra la ruta `/users`, se registrará también la ruta `/users/`
        y viceversa.
        """

        if path.endswith('/'):
            path = path[:-1]

        alternate_path = path + '/'

        if (self.prefix + path) != '':
            super().add_api_route(
                path,
                endpoint,
                response_model=response_model,
                status_code=status_code,
                tags=tags,
                dependencies=dependencies,
                summary=summary,
                description=description,
                response_description=response_description,
                responses=responses,
                deprecated=deprecated,
                methods=methods,
                operation_id=operation_id,
                response_model_include=response_model_include,
                response_model_exclude=response_model_exclude,
                response_model_by_alias=response_model_by_alias,
                response_model_exclude_unset=response_model_exclude_unset,
                response_model_exclude_defaults=response_model_exclude_defaults,
                response_model_exclude_none=response_model_exclude_none,
                include_in_schema=include_in_schema,
                response_class=response_class,
                name=name,
                route_class_override=route_class_override,
                callbacks=callbacks,
                openapi_extra=openapi_extra,
                generate_unique_id_function=generate_unique_id_function,
            )

        if (self.prefix + alternate_path) != '':
            super().add_api_route(
                alternate_path,
                endpoint,
                response_model=response_model,
                status_code=status_code,
                tags=tags,
                dependencies=dependencies,
                summary=summary,
                description=description,
                response_description=response_description,
                responses=responses,
                deprecated=deprecated,
                methods=methods,
                operation_id=operation_id,
                response_model_include=response_model_include,
                response_model_exclude=response_model_exclude,
                response_model_by_alias=response_model_by_alias,
                response_model_exclude_unset=response_model_exclude_unset,
                response_model_exclude_defaults=response_model_exclude_defaults,
                response_model_exclude_none=response_model_exclude_none,
                include_in_schema=False,
                response_class=response_class,
                name=name,
                route_class_override=route_class_override,
                callbacks=callbacks,
                openapi_extra=openapi_extra,
                generate_unique_id_function=generate_unique_id_function,
            )



from typing import Callable, TypeVar

try:
    from typing import Unpack
except ImportError:
    from typing_extensions import Unpack

from ..core.module import Module
from ..decorators._common import meta_decorator
from ..metadata.types.module_meta import ModuleMeta
from .dicts.module_dict import ModuleMetaDict

Class = TypeVar('Class', bound=type)


def module(**options: Unpack[ModuleMetaDict]) -> Callable:
    """🐀 ⇝ decorator that marks a class as a module"""
    return meta_decorator(meta=options, base=Module, meta_type=ModuleMeta, singleton=True)


def mod(**options: Unpack[ModuleMetaDict]) -> Callable:
    """🐀 ⇝ decorator that marks a class as a module"""
    return module(**options)


def domain(**options: Unpack[ModuleMetaDict]) -> Callable:
    """🐀 ⇝ decorator that marks a class as a domain module"""
    return module(**options)


def dom(**options: Unpack[ModuleMetaDict]) -> Callable:
    """🐀 ⇝ decorator that marks a class as a domain module"""
    return module(**options)



from enum import Enum
from typing import Union


class HTTPStatusEnum(Enum):
    HTTP_100_CONTINUE = 100, 'Continue'
    HTTP_101_SWITCHING_PROTOCOLS = 101, 'Switching Protocols'
    HTTP_102_PROCESSING = 102, 'Processing'
    HTTP_103_EARLY_HINTS = 103, 'Early Hints'
    HTTP_200_OK = 200, 'OK'
    HTTP_201_CREATED = 201, 'Created'
    HTTP_202_ACCEPTED = 202, 'Accepted'
    HTTP_203_NON_AUTHORITATIVE_INFORMATION = 203, 'Non-Authoritative Information'
    HTTP_204_NO_CONTENT = 204, 'No Content'
    HTTP_205_RESET_CONTENT = 205, 'Reset Content'
    HTTP_206_PARTIAL_CONTENT = 206, 'Partial Content'
    HTTP_207_MULTI_STATUS = 207, 'Multi-Status'
    HTTP_208_ALREADY_REPORTED = 208, 'Already Reported'
    HTTP_226_IM_USED = 226, 'IM Used'
    HTTP_300_MULTIPLE_CHOICES = 300, 'Multiple Choices'
    HTTP_301_MOVED_PERMANENTLY = 301, 'Moved Permanently'
    HTTP_302_FOUND = 302, 'Found'
    HTTP_303_SEE_OTHER = 303, 'See Other'
    HTTP_304_NOT_MODIFIED = 304, 'Not Modified'
    HTTP_305_USE_PROXY = 305, 'Use Proxy'
    HTTP_307_TEMPORARY_REDIRECT = 307, 'Temporary Redirect'
    HTTP_308_PERMANENT_REDIRECT = 308, 'Permanent Redirect'
    HTTP_400_BAD_REQUEST = 400, 'Bad Request'
    HTTP_401_UNAUTHORIZED = 401, 'Unauthorized'
    HTTP_402_PAYMENT_REQUIRED = 402, 'Payment Required'
    HTTP_403_FORBIDDEN = 403, 'Forbidden'
    HTTP_404_NOT_FOUND = 404, 'Not Found'
    HTTP_405_METHOD_NOT_ALLOWED = 405, 'Method Not Allowed'
    HTTP_406_NOT_ACCEPTABLE = 406, 'Not Acceptable'
    HTTP_407_PROXY_AUTHENTICATION_REQUIRED = 407, 'Proxy Authentication Required'
    HTTP_408_REQUEST_TIMEOUT = 408, 'Request Timeout'
    HTTP_409_CONFLICT = 409, 'Conflict'
    HTTP_410_GONE = 410, 'Gone'
    HTTP_411_LENGTH_REQUIRED = 411, 'Length Required'
    HTTP_412_PRECONDITION_FAILED = 412, 'Precondition Failed'
    HTTP_413_REQUEST_ENTITY_TOO_LARGE = 413, 'Request Entity Too Large'
    HTTP_414_REQUEST_URI_TOO_LONG = 414, 'Request-URI Too Long'
    HTTP_415_UNSUPPORTED_MEDIA_TYPE = 415, 'Unsupported Media Type'
    HTTP_416_REQUESTED_RANGE_NOT_SATISFIABLE = 416, 'Requested Range Not Satisfiable'
    HTTP_417_EXPECTATION_FAILED = 417, 'Expectation Failed'
    HTTP_418_IM_A_TEAPOT = 418, "I'm a teapot"
    HTTP_421_MISDIRECTED_REQUEST = 421, 'Misdirected Request'
    HTTP_422_UNPROCESSABLE_ENTITY = 422, 'Unprocessable Entity'
    HTTP_423_LOCKED = 423, 'Locked'
    HTTP_424_FAILED_DEPENDENCY = 424, 'Failed Dependency'
    HTTP_425_TOO_EARLY = 425, 'Too Early'
    HTTP_426_UPGRADE_REQUIRED = 426, 'Upgrade Required'
    HTTP_428_PRECONDITION_REQUIRED = 428, 'Precondition Required'
    HTTP_429_TOO_MANY_REQUESTS = 429, 'Too Many Requests'
    HTTP_431_REQUEST_HEADER_FIELDS_TOO_LARGE = 431, 'Request Header Fields Too Large'
    HTTP_451_UNAVAILABLE_FOR_LEGAL_REASONS = 451, 'Unavailable For Legal Reasons'
    HTTP_500_INTERNAL_SERVER_ERROR = 500, 'Internal Server Error'
    HTTP_501_NOT_IMPLEMENTED = 501, 'Not Implemented'
    HTTP_502_BAD_GATEWAY = 502, 'Bad Gateway'
    HTTP_503_SERVICE_UNAVAILABLE = 503, 'Service Unavailable'
    HTTP_504_GATEWAY_TIMEOUT = 504, 'Gateway Timeout'
    HTTP_505_HTTP_VERSION_NOT_SUPPORTED = 505, 'HTTP Version Not Supported'
    HTTP_506_VARIANT_ALSO_NEGOTIATES = 506, 'Variant Also Negotiates'
    HTTP_507_INSUFFICIENT_STORAGE = 507, 'Insufficient Storage'
    HTTP_508_LOOP_DETECTED = 508, 'Loop Detected'
    HTTP_510_NOT_EXTENDED = 510, 'Not Extended'
    HTTP_511_NETWORK_AUTHENTICATION_REQUIRED = 511, 'Network Authentication Required'


class HttpStatus:
    def __init__(self, code: int, phrase: str):
        self.code = code
        self.phrase = phrase


def http_status(status_code: Union[int, HTTPStatusEnum]) -> HttpStatus:
    if isinstance(status_code, HTTPStatusEnum):
        code = status_code.value[0]
    else:
        code = status_code

    for status in HTTPStatusEnum:
        if status.value[0] == code:
            return HttpStatus(status.value[0], status.value[1])

    # default to 500
    return http_status(HTTPStatusEnum.HTTP_500_INTERNAL_SERVER_ERROR)



from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Callable, List, Sequence, Union

from starlette.types import Lifespan

from ._meta import Meta, PestType


@dataclass
class ControllerMeta(Meta):
    meta_type: PestType = field(default=PestType.CONTROLLER, init=False, metadata={'expose': False})
    prefix: str = field(metadata={'expose': False})
    tags: Union[List[Union[str, Enum]], None]
    '''🐀 ⇝ tags of the controller'''
    redirect_slashes: Union[bool, None]
    '''🐀 ⇝ redirect slashes?'''
    on_startup: Union[Sequence[Callable[[], Any]], None]
    '''🐀 ⇝ on startup events'''
    on_shutdown: Union[Sequence[Callable[[], Any]], None]
    '''🐀 ⇝ on shutdown events'''
    lifespan: Union[Lifespan[Any], None] = field(metadata={'expose': False})
    '''🐀 ⇝ lifespan of the controller'''
    deprecated: Union[bool, None]
    '''🐀 ⇝ is the controller deprecated?'''
    include_in_schema: Union[bool, None]
    '''🐀 ⇝ include in schema?'''



from enum import Enum
from typing import (
    Any,
    Callable,
    Coroutine,
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
    cast,
)

try:
    from typing import Unpack
except ImportError:
    from typing_extensions import Unpack

from dij import ActivationScope
from fastapi import FastAPI, Response, routing
from fastapi.datastructures import Default, DefaultPlaceholder
from fastapi.exceptions import RequestValidationError, WebSocketRequestValidationError
from fastapi.params import Depends
from fastapi.responses import JSONResponse
from fastapi.types import DecoratedCallable, IncEx
from fastapi.utils import generate_unique_id
from pydantic import ValidationError
from starlette.exceptions import HTTPException
from starlette.middleware import Middleware
from starlette.middleware.errors import ServerErrorMiddleware
from starlette.middleware.exceptions import ExceptionMiddleware
from starlette.routing import BaseRoute
from starlette.types import ASGIApp

from pest.logging import log
from pest.middleware.types import CorsOptions

from ..exceptions import handle
from ..metadata.types.module_meta import InjectionToken
from ..middleware.base import (
    PestBaseHTTPMiddleware,
    PestMiddlwareCallback,
)
from ..middleware.di import di_scope_middleware
from ..middleware.types import MiddlewareDef
from .module import Module, T
from .types.fastapi_params import FastAPIParams


def root_module(app: 'PestApplication') -> Module:
    return app.__pest_module__


class PestApplication(FastAPI):
    """🐀 ⇝ what a pest!"""

    def __init__(
        self, module: Module, middleware: MiddlewareDef, **kwargs: Unpack[FastAPIParams]
    ) -> None:
        super().__init__(**kwargs)
        self.__pest_module__ = module

        self.user_middleware: List[Middleware] = (
            []
            if middleware is None
            else [
                (
                    middleware
                    if isinstance(middleware, Middleware)
                    else Middleware(
                        PestBaseHTTPMiddleware,
                        dispatch=cast(PestMiddlwareCallback, middleware),
                        parent_module=module,
                    )
                )
                for middleware in middleware
            ]
        )

        self.add_exception_handlers([
            (HTTPException, handle.http),
            (ValidationError, handle.request_validation),
            (RequestValidationError, handle.request_validation),
            (WebSocketRequestValidationError, handle.websocket_request_validation),
            # for everything else, there's Mastercard (or was it Bancard? 🤔)
            (Exception, handle.the_rest),
        ])

    def add_exception_handlers(
        self, handlers: List[Tuple[Union[int, Type[Exception]], Callable]]
    ) -> None:
        for error, handler in handlers:
            self.add_exception_handler(error, handler)

    def __str__(self) -> str:
        return str(root_module(self))

    def resolve(self, token: InjectionToken[T], scope: Union[ActivationScope, None] = None) -> T:
        return root_module(self).get(token, scope)

    async def aresolve(
        self, token: InjectionToken[T], scope: Union[ActivationScope, None] = None
    ) -> T:
        return await root_module(self).aget(token, scope)

    def can_provide(self, token: InjectionToken[T]) -> bool:
        return root_module(self).can_provide(token)

    def add_api_route(
        self,
        path: str,
        endpoint: Callable[..., Coroutine[Any, Any, Response]],
        *,
        response_model: Any = Default(None),
        status_code: Optional[int] = None,
        tags: Optional[List[Union[str, Enum]]] = None,
        dependencies: Optional[Sequence[Depends]] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        response_description: str = 'Successful Response',
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        deprecated: Optional[bool] = None,
        methods: Optional[List[str]] = None,
        operation_id: Optional[str] = None,
        response_model_include: Optional[IncEx] = None,
        response_model_exclude: Optional[IncEx] = None,
        response_model_by_alias: bool = True,
        response_model_exclude_unset: bool = False,
        response_model_exclude_defaults: bool = False,
        response_model_exclude_none: bool = False,
        include_in_schema: bool = True,
        response_class: Union[Type[Response], DefaultPlaceholder] = Default(JSONResponse),
        name: Optional[str] = None,
        openapi_extra: Optional[Dict[str, Any]] = None,
        generate_unique_id_function: Callable[[routing.APIRoute], str] = Default(
            generate_unique_id
        ),
        **kwargs: Any,
    ) -> None:
        self.router.add_api_route(
            path,
            endpoint=endpoint,
            response_model=response_model,
            status_code=status_code,
            tags=tags,
            dependencies=dependencies,
            summary=summary,
            description=description,
            response_description=response_description,
            responses=responses,
            deprecated=deprecated,
            methods=methods,
            operation_id=operation_id,
            response_model_include=response_model_include,
            response_model_exclude=response_model_exclude,
            response_model_by_alias=response_model_by_alias,
            response_model_exclude_unset=response_model_exclude_unset,
            response_model_exclude_defaults=response_model_exclude_defaults,
            response_model_exclude_none=response_model_exclude_none,
            include_in_schema=include_in_schema,
            response_class=response_class,
            name=name,
            openapi_extra=openapi_extra,
            generate_unique_id_function=generate_unique_id_function,
            **kwargs,
        )

    def include_router(
        self,
        router: routing.APIRouter,
        *,
        prefix: str = '',
        tags: Optional[List[Union[str, Enum]]] = None,
        dependencies: Optional[Sequence[Depends]] = None,
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        deprecated: Optional[bool] = None,
        include_in_schema: bool = True,
        default_response_class: Type[Response] = Default(JSONResponse),
        callbacks: Optional[List[BaseRoute]] = None,
        generate_unique_id_function: Callable[[routing.APIRoute], str] = Default(
            generate_unique_id
        ),
    ) -> None:
        self.router.include_router(
            router,
            prefix=prefix,
            tags=tags,
            dependencies=dependencies,
            responses=responses,
            deprecated=deprecated,
            include_in_schema=include_in_schema,
            default_response_class=default_response_class,
            callbacks=callbacks,
            generate_unique_id_function=generate_unique_id_function,
        )

    def middleware(self, middleware_type: str) -> Callable[[DecoratedCallable], DecoratedCallable]:
        """
        Add a middleware to the application.

        Read more about it in the
        [FastAPI docs for Middleware](https://fastapi.tiangolo.com/tutorial/middleware/).

        ## Example

        ```python
        import time

        from fastapi import FastAPI, Request

        app = FastAPI()


        @app.middleware("http")
        async def add_process_time_header(request: Request, call_next):
            start_time = time.time()
            response = await call_next(request)
            process_time = time.time() - start_time
            response.headers["X-Process-Time"] = str(process_time)
            return response
        ```
        """

        def decorator(func: DecoratedCallable) -> DecoratedCallable:
            self.add_middleware(
                PestBaseHTTPMiddleware, dispatch=func, parent_module=root_module(self)
            )
            return func

        return decorator

    def enable_cors(self, **opts: Unpack[CorsOptions]) -> None:
        from starlette.middleware.cors import CORSMiddleware

        from ..middleware.types import DEFAULT_CORS_OPTIONS

        log.debug(f'Enabling CORS with options: {DEFAULT_CORS_OPTIONS | opts}')

        self.add_middleware(CORSMiddleware, **opts)

    def add_middleware(self, middleware_class: type, *args: Any, **kwargs: Any) -> None:
        super().add_middleware(middleware_class, *args, **kwargs)

    def build_middleware_stack(self) -> ASGIApp:
        # Duplicate/override from FastAPI to add the di_scope_middleware, which
        # is required for Pest's DI to work. We need it to run before any other
        # user-defined middleware, as it is responsible for injecting the
        # the per-request scope into the DI container, that might be needed
        # by other middlewares to be able to correctly resolve their dependencies.
        debug = self.debug
        error_handler = None
        exception_handlers = {}

        for key, value in self.exception_handlers.items():
            if key in (500, Exception):
                error_handler = value
            else:
                exception_handlers[key] = value

        di_scope_mw = [
            Middleware(
                PestBaseHTTPMiddleware,
                dispatch=di_scope_middleware,
                parent_module=root_module(self),
            )
        ]

        middleware = (
            [Middleware(ServerErrorMiddleware, handler=error_handler, debug=debug)]
            + di_scope_mw  # <--- this is the only difference
            + self.user_middleware
            + [Middleware(ExceptionMiddleware, handlers=exception_handlers, debug=debug)]
        )

        app = self.router
        for cls, args, kwargs in reversed(middleware):
            app = cls(app=app, *args, **kwargs)
        return app



from typing import Callable, Type, TypeVar, Union, cast

T = TypeVar('T')


class _Inject:
    """
    Class used to inject dependencies from the `module`'s container.
    @internal you should not use this class directly, use `inject` instead.
    """

    def __init__(self, token: Union[Type, None, Callable]) -> None:
        self.token = token


def inject(token: Union[None, Type[T], Callable] = None) -> T:
    """
    🐀 ⇝ marks a parameter as a dependency to be injected by `pest` 💉

    Similar to `fastapi.Depends` but this one is used to inject dependencies
    from the `module`'s private container, respecting encapsulation.

    #### Example

    ```py
    @get('/{user_id}')
    def get_user(self, user_id: str, svc: Annotated[Service, inject]):
        return svc.get_user(user_id)
    ```

    You can also use it without the `Annotated` annotation:

    ```py
    svc: Service = inject()

    # or even like this if you don't care about type hints
    svc = inject(Service)
    ```

    If you use it in a functional middleware, you can use it like this:

    ```py
    async def mw(request: Request, call_next: CallNext, svc: Service = inject()):
        # do something with svc
        return await call_next(request)
    ```
    """
    return cast(T, _Inject(token))



from typing import Any, Dict, List, Union

from pydantic import BaseModel, Field
from starlette.exceptions import HTTPException

from ...utils.functions import dump_model
from .status import http_status


class ExceptionResponse(BaseModel):
    """Error response model"""

    code: int = Field(..., description='HTTP status code')
    error: Union[str, None] = Field(default=None, description='HTTP status phrase')
    message: Union[List[str], str, None] = Field(default=None, description='Error message')

    @staticmethod
    def example(code: int) -> 'ExceptionResponse':
        status = http_status(code)

        return ExceptionResponse(
            code=status.code,
            error=status.phrase,
            message='Detailed error message',
        )


def exc_response(*codes: int) -> Dict[Union[int, str], Dict[str, Any]]:
    """🐀 ⇝ Generate a dict of error responses for the given status codes for use in OpenAPI docs"""

    responses = {}

    for code in codes:
        example = ExceptionResponse.example(code)
        responses[code] = {
            'description': example.error,
            'model': ExceptionResponse,
            'content': {
                'application/json': {
                    'example': example.model_dump(),
                },
            },
        }

    return responses


class PestHTTPException(HTTPException):
    """🐀 ⇝ base class for all pest HTTP exceptions"""

    def __init__(
        self,
        status_code: int,
        detail: Any = None,
        headers: Union[Dict[str, Any], None] = None,
    ) -> None:
        super().__init__(
            status_code=status_code,
            detail=detail,
            headers=headers,
        )

    @property
    def __dict__(self) -> Dict[str, Any]:
        status = http_status(self.status_code)

        return dump_model(
            ExceptionResponse(code=status.code, error=status.phrase, message=self.detail)
        )


# region: HTTP 4xx errors
class BadRequestException(PestHTTPException):
    """🐀 ⇝ `400` Bad Request

    The server cannot or will not process the request due to something that is perceived to be
    a client error (e.g., malformed request syntax, invalid request message framing, or deceptive
    request routing).

    [› more](https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/400)
    """

    def __init__(
        self,
        detail: Any = None,
        headers: Union[Dict[str, Any], None] = None,
    ) -> None:
        """🐀 ⇝ `400` Bad Request"""
        super().__init__(
            status_code=400,
            detail=detail,
            headers=headers,
        )


class UnauthorizedException(PestHTTPException):
    """🐀 ⇝ `401` Unauthorized

    Similar to `403 Forbidden`, but specifically for use when authentication is required and has
    failed or has not yet been provided. It indicates that the client must authenticate
    itself to get the requested response.

    [› more](https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/401)
    """

    def __init__(
        self,
        detail: Any = None,
        headers: Union[Dict[str, Any], None] = None,
    ) -> None:
        """🐀 ⇝ `401` Unauthorized"""
        super().__init__(
            status_code=401,
            detail=detail,
            headers=headers,
        )


class ForbiddenException(PestHTTPException):
    """🐀 ⇝ `403` Forbidden

    The client does not have access rights to the content; that is, it is unauthorized,
    so the server is refusing to give the requested resource. Unlike `401`, the client's
    identity is known to the server.

    [› more](https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/403)
    """

    def __init__(
        self,
        detail: Any = None,
        headers: Union[Dict[str, Any], None] = None,
    ) -> None:
        """🐀 ⇝ `403` Forbidden"""
        super().__init__(
            status_code=403,
            detail=detail,
            headers=headers,
        )


class NotFoundException(PestHTTPException):
    """🐀 ⇝ `404` Not Found

    The requested resource could not be found but may be available in the future.

    [› more](https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/404)
    """

    def __init__(
        self,
        detail: Any = None,
        headers: Union[Dict[str, Any], None] = None,
    ) -> None:
        """🐀 ⇝ `404` Not Found"""
        super().__init__(
            status_code=404,
            detail=detail,
            headers=headers,
        )


class MethodNotAllowedException(PestHTTPException):
    """🐀 ⇝ `405` Method Not Allowed

    A request method is not supported for the requested resource; for example, an API endpoint
    may not support `DELETE` requests.

    [› more](https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/405)
    """

    def __init__(
        self,
        detail: Any = None,
        headers: Union[Dict[str, Any], None] = None,
    ) -> None:
        """🐀 ⇝ `405` Method Not Allowed"""
        super().__init__(
            status_code=405,
            detail=detail,
            headers=headers,
        )


class NotAcceptableException(PestHTTPException):
    """🐀 ⇝ `406` Not Acceptable

    The requested resource is capable of generating only content not acceptable according to the
    Accept headers sent in the request.

    [› more](https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/406)
    """

    def __init__(
        self,
        detail: Any = None,
        headers: Union[Dict[str, Any], None] = None,
    ) -> None:
        """🐀 ⇝ `406` Not Acceptable"""
        super().__init__(
            status_code=406,
            detail=detail,
            headers=headers,
        )


class ProxyAuthenticationRequiredException(PestHTTPException):
    """🐀 ⇝ `407` Proxy Authentication Required

    This is similar to 401 Unauthorized but authentication is needed to be done by a proxy.

    [› more](https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/407)
    """

    def __init__(
        self,
        detail: Any = None,
        headers: Union[Dict[str, Any], None] = None,
        headers_: Union[Dict[str, Any], None] = None,
    ) -> None:
        """🐀 ⇝ `407` Proxy Authentication Required"""
        super().__init__(
            status_code=407,
            detail=detail,
            headers=headers,
        )
        self.headers = headers_


class RequestTimeoutException(PestHTTPException):
    """🐀 ⇝ `408` Request Timeout

    The server timed out waiting for the request.

    [› more](https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/408)
    """

    def __init__(
        self,
        detail: Any = None,
        headers: Union[Dict[str, Any], None] = None,
    ) -> None:
        """🐀 ⇝ `408` Request Timeout"""
        super().__init__(
            status_code=408,
            detail=detail,
            headers=headers,
        )


class ConflictException(PestHTTPException):
    """🐀 ⇝ `409` Conflict

    Indicates that the request could not be processed because of conflict in the request, such as
    an attempt to modify a resource in a way that would conflict with its current state.

    [› more](https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/409)
    """

    def __init__(
        self,
        detail: Any = None,
        headers: Union[Dict[str, Any], None] = None,
    ) -> None:
        """🐀 ⇝ `409` Conflict"""
        super().__init__(
            status_code=409,
            detail=detail,
            headers=headers,
        )


class GoneException(PestHTTPException):
    """🐀 ⇝ `410` Gone

    This response is sent when the requested content has been permanently deleted from server.

    [› more](https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/410)
    """

    def __init__(
        self,
        detail: Any = None,
        headers: Union[Dict[str, Any], None] = None,
    ) -> None:
        """🐀 ⇝ `410` Gone"""
        super().__init__(
            status_code=410,
            detail=detail,
            headers=headers,
        )


class PreconditionFailedException(PestHTTPException):
    """🐀 ⇝ `412` Precondition Failed

    The client has indicated preconditions in its headers which the server does not meet.

    [› more](https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/412)
    """

    def __init__(
        self,
        detail: Any = None,
        headers: Union[Dict[str, Any], None] = None,
    ) -> None:
        """🐀 ⇝ `412` Precondition Failed"""
        super().__init__(
            status_code=412,
            detail=detail,
            headers=headers,
        )


class PayloadTooLargeException(PestHTTPException):
    """🐀 ⇝ `413` Payload Too Large

    Request entity is larger than limits defined by server. The server might close the
    connection or return an Retry-After header field.

    [› more](https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/413)
    """

    def __init__(
        self,
        detail: Any = None,
        headers: Union[Dict[str, Any], None] = None,
    ) -> None:
        """🐀 ⇝ `413` Payload Too Large"""
        super().__init__(
            status_code=413,
            detail=detail,
            headers=headers,
        )


class UnsupportedMediaTypeException(PestHTTPException):
    """🐀 ⇝ `415` Unsupported Media Type

    The media format of the requested data is not supported by the server, so the
    server is rejecting the request

    [› more](https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/415)
    """

    def __init__(
        self,
        detail: Any = None,
        headers: Union[Dict[str, Any], None] = None,
        media_type: Union[str, None] = None,
    ) -> None:
        """🐀 ⇝ `415` Unsupported Media Type"""
        super().__init__(
            status_code=415,
            detail=detail,
            headers=headers,
        )
        self.media_type = media_type


class ImATeapotException(PestHTTPException):
    """🐀 ⇝ `418` I'm a teapot

    The server was asked to brew coffee but it can't because it is, permanently, a teapot.

    🤪

    [› more](https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/418)
    """

    def __init__(
        self,
        detail: Any = None,
        headers: Union[Dict[str, Any], None] = None,
    ) -> None:
        """🐀 ⇝ `418` I'm a teapot"""
        super().__init__(
            status_code=418,
            detail=detail,
            headers=headers,
        )


class UnprocessableEntityException(PestHTTPException):
    """🐀 ⇝ `422` Unprocessable Entity

    The request was well-formed but was unable to be followed due to semantic errors.

    [› more](https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/422)
    """

    def __init__(
        self,
        detail: Any = None,
        headers: Union[Dict[str, Any], None] = None,
    ) -> None:
        """🐀 ⇝ `422` Unprocessable Entity"""
        super().__init__(
            status_code=422,
            detail=detail,
            headers=headers,
        )


# endregion


# region: HTTP 5xx errors
class InternalServerErrorException(PestHTTPException):
    """🐀 ⇝ `500` Internal Server Error

    A generic error message, given when an unexpected condition was encountered and no more
    specific message is suitable or available.

    [› more](https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/500)
    """

    def __init__(
        self,
        detail: Any = None,
        headers: Union[Dict[str, Any], None] = None,
    ) -> None:
        """🐀 ⇝ `500` Internal Server Error"""
        super().__init__(
            status_code=500,
            detail=detail,
            headers=headers,
        )


class NotImplementedException(PestHTTPException):
    """🐀 ⇝ `501` Not Implemented

    The request method is not supported by the server and cannot be handled.

    [› more](https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/501)
    """

    def __init__(
        self,
        detail: Any = None,
        headers: Union[Dict[str, Any], None] = None,
    ) -> None:
        """🐀 ⇝ `501` Not Implemented"""
        super().__init__(
            status_code=501,
            detail=detail,
            headers=headers,
        )


class BadGatewayException(PestHTTPException):
    """🐀 ⇝ `502` Bad Gateway

    This error response means that the server, while working as a gateway to get a response
    needed to handle the request, got an invalid response.

    [› more](https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/502)
    """

    def __init__(
        self,
        detail: Any = None,
        headers: Union[Dict[str, Any], None] = None,
    ) -> None:
        """🐀 ⇝ `502` Bad Gateway"""
        super().__init__(
            status_code=502,
            detail=detail,
            headers=headers,
        )


class ServiceUnavailableException(PestHTTPException):
    """🐀 ⇝ `503` Service Unavailable

    The server is not ready to handle the request for some reason.

    [› more](https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/503)
    """

    def __init__(
        self,
        detail: Any = None,
        headers: Union[Dict[str, Any], None] = None,
    ) -> None:
        """🐀 ⇝ `503` Service Unavailable"""
        super().__init__(
            status_code=503,
            detail=detail,
            headers=headers,
        )


class GatewayTimeoutException(PestHTTPException):
    """🐀 ⇝ `504` Gateway Timeout

    This error response is given when the server is acting as a gateway and cannot get
    a response in time.

    [› more](https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/504)
    """

    def __init__(
        self,
        detail: Any = None,
        headers: Union[Dict[str, Any], None] = None,
    ) -> None:
        """🐀 ⇝ `504` Gateway Timeout"""
        super().__init__(
            status_code=504,
            detail=detail,
            headers=headers,
        )


class HttpVersionNotSupportedException(PestHTTPException):
    """🐀 ⇝ `505` HTTP Version Not Supported

    The HTTP version used in the request is not supported by the server.

    [› more](https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/505)
    """

    def __init__(
        self,
        detail: Any = None,
        headers: Union[Dict[str, Any], None] = None,
    ) -> None:
        """🐀 ⇝ `505` HTTP Version Not Supported"""
        super().__init__(
            status_code=505,
            detail=detail,
            headers=headers,
        )


# endregion



from dataclasses import asdict, fields, is_dataclass
from typing import Any, Callable, Dict, List, Mapping, Type, TypeVar, Union, cast

from dacite import Config, from_dict

from ..exceptions.base.pest import PestException
from ..utils.functions import drop_keys, keep_keys
from ..utils.protocols import DataclassInstance
from .types._meta import Meta

META_KEY = '__pest__'


DataType = TypeVar('DataType', bound=Union[Dict[str, Any], dict, Meta, DataclassInstance])
GenericValue = TypeVar('GenericValue')


def get_meta(
    target: Union[Callable[..., Any], type, object],
    output_type: Type[DataType] = Dict[str, Any],
    *,
    raise_error: bool = True,
    clean: bool = False,
    keep: Union[List[str], None] = None,
    drop: Union[List[str], None] = None,
) -> DataType:
    """🐀 ⇝ get `metadata` from a `callable`
    #### Params
    - target: target object, type or function
    - type: return type (will create an instance if it's a `dataclass`)
    - raise_error: wether to raise an error if no metadata was found in the target
    - clean: wether to clean up the resulting object (will remove `meta_type` by default)
    - keep: if `clean == True`, will only keep the key the keys provided here
    - drop: if `clean == True`, will drop all keys provided here and keep the rest.
    """

    if not hasattr(target, META_KEY):
        if raise_error:
            raise PestException(f'No metadata for {target}')
        return cast(DataType, {})

    meta = getattr(target, META_KEY)

    # clean up the metadata
    if clean:
        if keep is not None and len(keep) > 0:
            meta = keep_keys(meta, keep)

        if drop is not None and len(drop) > 0:
            meta = drop_keys(meta, drop)

        if not keep and not drop:
            meta = drop_keys(meta, ['meta_type'])

    if is_dataclass(output_type):
        return cast(output_type, from_dict(output_type, meta, config=Config(check_types=False)))  # type: ignore

    return cast(output_type, meta)  # type: ignore


def get_meta_value(
    callable: Callable[..., Any],
    key: str,
    default: Any = None,
    *,
    type: Type[GenericValue] = Type[Any],
) -> GenericValue:
    """🐀 ⇝ get pest metadata `value` from a `callable` by `key`"""

    meta = get_meta(callable, raise_error=False)
    return cast(type, meta.get(key, default))  # type: ignore


def inject_metadata(
    callable: Callable[..., Any],
    metadata: Union[Meta, Mapping[Any, Any], None] = None,
    **kwargs: Any,
) -> None:
    """🐀 ⇝ initialize pest `metadata` for a `callable`

    callable: The callable to initialize metadata for
    **kwargs: keyword arguments to initialize metadata with
    """

    if not hasattr(callable, META_KEY):
        setattr(callable, META_KEY, {})

    dict_meta = {}
    if metadata is not None:
        if not is_dataclass(metadata) and not isinstance(metadata, dict):
            raise PestException('metadata must be a dataclass or a dict')

        try:
            dict_meta: dict = (
                asdict(metadata)
                if is_dataclass(metadata) and not isinstance(metadata, type)
                else metadata if isinstance(metadata, dict) else {}
            )
        except Exception as e:
            if is_dataclass(metadata) and not isinstance(metadata, type):
                dict_meta = {}
                for field in fields(metadata):
                    dict_meta[field.name] = getattr(metadata, field.name)
            else:
                raise e

    meta = get_meta(callable)
    meta.update({**dict_meta, **kwargs})



from dataclasses import dataclass
from enum import Enum
from typing import Protocol


class PestType(str, Enum):
    MODULE = 'MODULE'
    INJECTABLE = 'INJECTABLE'
    CONTROLLER = 'CONTROLLER'
    HANDLER = 'HANDLER'


@dataclass
class Meta(Protocol):
    meta_type: PestType
    '''🐀 ⇝ type of the metadata'''



from dataclasses import dataclass, field
from typing import List, Type, Union

from ...core.controller import Controller
from ._meta import Meta, PestType
from .injectable_meta import (
    ClassProvider,
    ExistingProvider,
    FactoryProvider,
    InjectionToken,
    ValueProvider,
)

try:
    from typing import TypeAlias
except ImportError:
    from typing_extensions import TypeAlias


Class: TypeAlias = type
Provider: TypeAlias = Union[
    Class, 'ValueProvider', 'FactoryProvider', 'ExistingProvider', 'ClassProvider'
]


@dataclass
class ModuleMeta(Meta):
    meta_type: PestType = field(default=PestType.MODULE, init=False, metadata={'expose': False})

    imports: Union[List[type], None]
    '''🐀 ⇝ list of modules to be imported'''

    providers: Union[List[Provider], None]
    '''🐀 ⇝ list of providers to be registered'''

    exports: Union[List[InjectionToken], None]
    '''🐀 ⇝ list of providers to be exported'''

    controllers: Union[List[Type[Controller]], None]
    '''🐀 ⇝ list of controllers to be registered'''



from typing import Callable, Type, TypeVar

try:
    from typing import Unpack
except ImportError:
    from typing_extensions import Unpack


from pest.decorators._common import meta_decorator

from ..core.controller import Controller
from ..metadata.types.controller_meta import ControllerMeta
from .dicts.controller_dict import ControllerMetaDict

Class = TypeVar('Class', bound=type)


def controller(
    prefix: str, **options: Unpack[ControllerMetaDict]
) -> Callable[..., Type[Controller]]:
    """🐀 ⇝ decorator that marks a class as a `controller`"""
    controller_prefix: str | None = prefix

    if prefix == '/':
        controller_prefix = None

    return meta_decorator(
        meta_type=ControllerMeta,
        meta={
            'prefix': controller_prefix,
            **options,
        },
        base=Controller,
    )


def ctrl(prefix: str, **options: Unpack[ControllerMetaDict]) -> Callable:
    """🐀 ⇝ decorator that marks a class as a `controller`"""
    return controller(prefix, **options)


def router(prefix: str, **options: Unpack[ControllerMetaDict]) -> Callable:
    """🐀 ⇝ decorator that marks a class as a `router`"""
    return controller(prefix, **options)


def rtr(prefix: str, **options: Unpack[ControllerMetaDict]) -> Callable:
    """🐀 ⇝ decorator that marks a class as a `router`"""
    return controller(prefix, **options)


def api(prefix: str, **options: Unpack[ControllerMetaDict]) -> Callable:
    """🐀 ⇝ decorator that marks a class as a `router`"""
    return controller(prefix, **options)



from enum import Enum
from typing import Callable

try:
    from typing import Unpack
except ImportError:
    from typing_extensions import Unpack

from fastapi.types import DecoratedCallable

from pest.decorators._common import meta_decorator

from ..metadata.types.handler_meta import HandlerMeta
from .dicts.handler_dict import HandlerMetaDict


class HttpMethod(str, Enum):
    GET = 'GET'
    HEAD = 'HEAD'
    POST = 'POST'
    PUT = 'PUT'
    DELETE = 'DELETE'
    CONNECT = 'CONNECT'
    OPTIONS = 'OPTIONS'
    TRACE = 'TRACE'
    PATCH = 'PATCH'


def __make(
    path: str, method: HttpMethod, **options: Unpack[HandlerMetaDict]
) -> Callable[[DecoratedCallable], DecoratedCallable]:
    """utility function to create a handler decorator"""
    return meta_decorator(
        meta_type=HandlerMeta,
        meta={
            'methods': [method],
            'path': path,
            **options,
        },
    )


def get(
    path: str, **options: Unpack[HandlerMetaDict]
) -> Callable[[DecoratedCallable], DecoratedCallable]:
    """🐀 ⇝ marks a function as a `GET` request handler"""
    return __make(path, HttpMethod.GET, **options)


def post(
    path: str, **options: Unpack[HandlerMetaDict]
) -> Callable[[DecoratedCallable], DecoratedCallable]:
    """🐀 ⇝ marks a function as a `POST` request handler"""
    return __make(path, HttpMethod.POST, **options)


def put(
    path: str, **options: Unpack[HandlerMetaDict]
) -> Callable[[DecoratedCallable], DecoratedCallable]:
    """🐀 ⇝ marks a function as a `PUT` request handler"""
    return __make(path, HttpMethod.PUT, **options)


def delete(
    path: str, **options: Unpack[HandlerMetaDict]
) -> Callable[[DecoratedCallable], DecoratedCallable]:
    """🐀 ⇝ marks a function as a `DELETE` request handler"""
    return __make(path, HttpMethod.DELETE, **options)


def patch(
    path: str, **options: Unpack[HandlerMetaDict]
) -> Callable[[DecoratedCallable], DecoratedCallable]:
    """🐀 ⇝ marks a function as a `PATCH` request handler"""
    return __make(path, HttpMethod.PATCH, **options)


def options(
    path: str, **options: Unpack[HandlerMetaDict]
) -> Callable[[DecoratedCallable], DecoratedCallable]:
    """🐀 ⇝ marks a function as a `OPTIONS` request handler"""
    return __make(path, HttpMethod.OPTIONS, **options)


def head(
    path: str, **options: Unpack[HandlerMetaDict]
) -> Callable[[DecoratedCallable], DecoratedCallable]:
    """🐀 ⇝ marks a function as a `HEAD` request handler"""
    return __make(path, HttpMethod.HEAD, **options)


def trace(
    path: str, **options: Unpack[HandlerMetaDict]
) -> Callable[[DecoratedCallable], DecoratedCallable]:
    """🐀 ⇝ marks a function as a `TRACE` request handler"""
    return __make(path, HttpMethod.TRACE, **options)



from typing import Union

from ..logging import LoggingOptions


def setup(
    logging: Union[LoggingOptions, None] = None,
) -> None:
    """Initializes stuff that needs to be done before the application is created"""
    __setup_logging(logging)


def __setup_logging(logging: Union[LoggingOptions, None]) -> None:
    if logging is not None:
        # loguru needs to be installed in order to be able to import this
        from ..logging.loguru import Loguru

        Loguru.setup(**logging)



from dataclasses import dataclass, field
from typing import Callable, Generic, Type, TypeVar, Union

from dij import ServiceLifeStyle

from ._meta import Meta, PestType

try:
    from typing import TypeAlias
except ImportError:
    from typing_extensions import TypeAlias

T = TypeVar('T')


Factory: TypeAlias = Callable[..., T]
Scope: TypeAlias = ServiceLifeStyle
InjectionToken: TypeAlias = Union[type, Type[T]]
Class: TypeAlias = type


@dataclass
class InjectableMeta(Meta):
    meta_type: PestType = field(default=PestType.INJECTABLE, init=False)


@dataclass
class ProviderBase:
    """🐀 ⇝ base class for all providers."""

    provide: InjectionToken
    '''🐀 ⇝ unique injection token'''


@dataclass
class ClassProvider(ProviderBase):
    """🐀 ⇝ defines a `class` type provider"""

    use_class: Class
    '''🐀 ⇝ type (class) of provider (type of the instance to be injected 💉)'''
    scope: Union[Scope, None] = None
    '''🐀 ⇝ scope of the provider''' ''


@dataclass
class ValueProvider(ProviderBase, Generic[T]):
    """🐀 ⇝ defines a `value` (singleton) type provider"""

    use_value: T
    '''🐀 ⇝ instance to be injected 💉'''


@dataclass
class SingletonProvider(ValueProvider, Generic[T]):
    """🐀 ⇝ defines a `singleton` (value) type provider"""

    pass


@dataclass
class FactoryProvider(ProviderBase, Generic[T]):
    """🐀 ⇝ defines a `factory` type provider"""

    use_factory: Factory[T]
    '''🐀 ⇝ factory function that returns an instance of the provider'''
    scope: Union[Scope, None] = None
    '''🐀 ⇝ scope of the provider'''


@dataclass
class ExistingProvider(ProviderBase, Generic[T]):
    """🐀 ⇝ defines an `existing` (aliased) type provider"""

    provide: str
    '''🐀 ⇝ unique injection token of the existing provider'''
    use_existing: InjectionToken
    '''🐀 ⇝ provider to be aliased by the injection token '''



from inspect import Signature, isclass, isfunction
from typing import (
    Any,
    Callable,
    Optional,
    Protocol,
    Tuple,
    Type,
    Union,
    cast,
    final,
    runtime_checkable,
)

try:
    from typing import TypeAlias, TypeGuard
except ImportError:
    from typing_extensions import TypeAlias, TypeGuard

from dij import ActivationScope
from starlette.middleware.base import BaseHTTPMiddleware, DispatchFunction, T
from starlette.middleware.base import RequestResponseEndpoint as CallNext
from starlette.requests import Request
from starlette.responses import Response
from starlette.types import ASGIApp

from ..core.module import Module
from ..metadata.types.module_meta import InjectionToken
from .di import scope_from

ProvideFn: TypeAlias = Callable[[InjectionToken[T], Optional[ActivationScope]], T]
UseFn: TypeAlias = Callable[[Request, CallNext], Response]


@runtime_checkable
class PestMiddlwareCallback(Protocol):
    async def __call__(
        self,
        request: Request,
        call_next: CallNext,
    ) -> Response: ...


@runtime_checkable
class PestMiddleware(Protocol):
    async def use(self, request: Request, call_next: CallNext) -> Response: ...

    @final
    async def __call__(
        self,
        request: Request,
        call_next: CallNext,
    ) -> Response:
        return await self.use(request, call_next)


class PestBaseHTTPMiddleware(BaseHTTPMiddleware):
    """🐀 ⇝ asgi middleware that receivs an injetor function and a dispatch funtion

    same as `starlette`'s `BaseHTTPMiddleware` but this one supports di injection
    """

    def __init__(
        self, app: ASGIApp, parent_module: Module, dispatch: PestMiddlwareCallback
    ) -> None:
        self.parent_module = parent_module
        super().__init__(app, dispatch=self.__dispatch_fn(dispatch))

    def __dispatch_fn(
        self, dispatch: Union[PestMiddlwareCallback, Type[PestMiddlwareCallback]]
    ) -> Union[DispatchFunction, None]:
        if dispatch is None:
            return dispatch

        if _is_class_pest_mw_callback(dispatch):
            self.parent_module.register(dispatch)

        async def wrapper(request: Request, call_next: CallNext) -> Response:
            scope = scope_from(request)
            dispatch_fn = cast(
                PestMiddlwareCallback,
                (
                    dispatch
                    if not isclass(dispatch)
                    else self.parent_module.get(dispatch, scope, fail_on_coroutine=False)
                ),
            )

            args, kwargs = await self.__resolve_dispatcher_args(dispatch_fn, scope)
            return await dispatch_fn(request, call_next, *args, **kwargs)

        return wrapper

    async def __resolve_dispatcher_args(
        self, function: DispatchFunction, scope: Union[ActivationScope, None]
    ) -> Tuple[tuple, dict]:
        signature = Signature.from_callable(function)
        parameters = signature.parameters
        args = []
        kwargs = {}

        arg_count = 0
        for name, param in parameters.items():
            if arg_count < 2:
                arg_count += 1
                continue

            if param.kind == param.POSITIONAL_ONLY or param.kind == param.POSITIONAL_OR_KEYWORD:
                args.append(await self.parent_module.aget(param.annotation, scope))

            elif param.kind == param.KEYWORD_ONLY:
                kwargs[name] = await self.parent_module.aget(param.annotation, scope)

        return tuple(args), kwargs


def _is_class_pest_mw_callback(obj: Any) -> TypeGuard[Type[PestMiddlwareCallback]]:
    """checks if an object is a **class** that respects the pest middleware callback protocol"""
    return _is_pest_mw_callback(obj) and isclass(obj)


def _is_pest_mw_callback(
    obj: Any,
) -> TypeGuard[Union[PestMiddlwareCallback, Type[PestMiddlwareCallback]]]:
    """checks if an object respects the pest middleware callback protocol"""
    respects_protocol = (
        isclass(obj)
        and issubclass(obj, PestMiddlwareCallback)
        or callable(obj)
        and isinstance(obj, PestMiddlwareCallback)
    )

    if not respects_protocol:
        return False

    # Check the function signature
    is_function = isfunction(obj)
    fn = obj if is_function else obj.__call__
    params = Signature.from_callable(fn).parameters
    args = list(params) if is_function else list(params)[1:]

    # Check if the first two arguments are 'request' and 'call_next'
    if args[:2] != ['request', 'call_next']:
        return False

    return True



from typing import Union

from ..core.application import PestApplication
from ..exceptions.http.http import ExceptionResponse
from ..middleware.types import CorsOptions
from ..utils.functions import model_schema


def setup(app: PestApplication, cors: Union[CorsOptions, None] = None) -> PestApplication:
    """Initializes stuff that needs the app to already exist to be able to setup"""
    __setup_cors(app, cors)
    __patch_open_api(app)

    return app


def __setup_cors(app: PestApplication, cors: Union[CorsOptions, None]) -> None:
    """Sets up CORS"""
    if cors is None:
        return

    app.enable_cors(**cors)


def __patch_open_api(app: PestApplication) -> None:
    """Patches the OpenAPI schema"""
    openapi_schema = app.openapi()
    if not openapi_schema:
        return

    for path in openapi_schema['paths']:
        for method in openapi_schema['paths'][path]:
            if openapi_schema['paths'][path][method]['responses'].get('422'):
                openapi_schema['paths'][path][method]['responses']['400'] = openapi_schema['paths'][
                    path
                ][method]['responses']['422']

                # change schema ref to ErrorReponse instead of HTTPValidationError
                openapi_schema['paths'][path][method]['responses']['400']['content'][
                    'application/json'
                ]['schema']['$ref'] = '#/components/schemas/ExceptionResponse'

                # set example
                openapi_schema['paths'][path][method]['responses']['400']['content'][
                    'application/json'
                ]['example'] = {
                    'code': 400,
                    'error': 'Bad Request',
                    'message': [
                        'error hint',
                    ],
                }

                # remove 422
                openapi_schema['paths'][path][method]['responses'].pop('422')

        components = openapi_schema.get('components')
        if components:
            # remove ValidationError and HTTPValidationError
            schemas = components.get('schemas')
            if schemas:
                if schemas.get('ValidationError'):
                    schemas.pop('ValidationError')
                if schemas.get('HTTPValidationError'):
                    schemas.pop('HTTPValidationError')
                if not schemas.get('ExceptionResponse'):
                    schemas['ExceptionResponse'] = model_schema(ExceptionResponse)

        app.openapi_schema = openapi_schema
